# -*- coding: utf-8 -*-

from logging import root
import os
from re import S 
import torch
import random
import h5py 
import pandas as pd
from scipy import ndimage
from torch.utils.data import Dataset
from torch.utils.data.sampler import Sampler

class H5DataSets(Dataset):
    """
    Dataset for loading images stored in h5 format. It generates 
    4D tensors with dimention order [C, D, H, W] for 3D images, and 
    3D tensors with dimention order [C, H, W] for 2D images
    """
    def __init__(self, root_dir, sample_list_name, transform = None):
        self.root_dir = root_dir 
        self.transform  = transform
        with open(sample_list_name, 'r') as f:
            lines = f.readlines()
            self.sample_list = [item.replace('\n', '') for item in lines]

    def __len__(self):
        return len(self.sample_list)

    def __getitem__(self, idx):
        sample_name = self.sample_list[idx]
        h5f = h5py.File(self.root_dir + '/' +  sample_name, 'r')
        image = h5f['image'][:]
        label = h5f['label'][:]
        sample = {'image': image, 'label': label}
        if self.transform:
            sample = self.transform(sample)
        # sample["idx"] = idx
        return sample
    
class TwoStreamBatchSampler(Sampler):
    """Iterate two sets of indices

    An 'epoch' is one iteration through the primary indices.
    During the epoch, the secondary indices are iterated through
    as many times as needed.
    """

    def __init__(self, primary_indices, secondary_indices, batch_size, secondary_batch_size):
        self.primary_indices = primary_indices
        self.secondary_indices = secondary_indices
        self.secondary_batch_size = secondary_batch_size
        self.primary_batch_size = batch_size - secondary_batch_size

        assert len(self.primary_indices) >= self.primary_batch_size > 0
        assert len(self.secondary_indices) >= self.secondary_batch_size > 0

    def __iter__(self):
        primary_iter = iterate_once(self.primary_indices)
        secondary_iter = iterate_eternally(self.secondary_indices)
        return (
            primary_batch + secondary_batch
            for (primary_batch, secondary_batch)
            in zip(grouper(primary_iter, self.primary_batch_size),
                   grouper(secondary_iter, self.secondary_batch_size))
        )

    def __len__(self):
        return len(self.primary_indices) // self.primary_batch_size


def iterate_once(iterable):
    return np.random.permutation(iterable)


def iterate_eternally(indices):
    def infinite_shuffles():
        while True:
            yield np.random.permutation(indices)
    return itertools.chain.from_iterable(infinite_shuffles())


def grouper(iterable, n):
    "Collect data into fixed-length chunks or blocks"
    # grouper('ABCDEFG', 3) --> ABC DEF"
    args = [iter(iterable)] * n
    return zip(*args)


if __name__ == "__main__":
    root_dir = "/home/guotai/disk2t/projects/semi_supervise/SSL4MIS/data/ACDC/data/slices"
    file_name = "/home/guotai/disk2t/projects/semi_supervise/slices.txt"
    dataset = H5DataSets(root_dir, file_name)
    train_loader = torch.utils.data.DataLoader(dataset, 
                batch_size = 4, shuffle=True, num_workers= 1)
    for sample in train_loader:
        image = sample['image']
        label = sample['label']
        print(image.shape, label.shape)
        print(image.min(), image.max(), label.max())


# -*- coding: utf-8 -*-
from __future__ import print_function, division
import logging
import torch
import numpy as np
from torch.optim import lr_scheduler
from pymic.loss.seg.util import get_soft_label
from pymic.loss.seg.util import reshape_prediction_and_ground_truth
from pymic.loss.seg.util import get_classwise_dice
from pymic.net_run_ssl.ssl_mt import SSLMeanTeacher
from pymic.util.ramps import get_rampup_ratio

class SSLUncertaintyAwareMeanTeacher(SSLMeanTeacher):
    """
    Uncertainty Aware Mean Teacher according to the following paper:
    Lequan Yu, Shujun Wang, Xiaomeng Li, Chi-Wing Fu, and Pheng-Ann Heng.
    Uncertainty-aware Self-ensembling Model for Semi-supervised 3D Left 
    Atrium Segmentation, MICCAI 2019.
    https://arxiv.org/abs/1907.07034 
    """
    def training(self):
        class_num   = self.config['network']['class_num']
        iter_valid  = self.config['training']['iter_valid']
        ssl_cfg     = self.config['semi_supervised_learning']
        iter_max     = self.config['training']['iter_max']
        rampup_start = ssl_cfg.get('rampup_start', 0)
        rampup_end   = ssl_cfg.get('rampup_end', iter_max)
        train_loss  = 0
        train_loss_sup = 0
        train_loss_reg = 0
        train_dice_list = []
        self.net.train()
        self.net_ema.to(self.device)
        for it in range(iter_valid):
            try:
                data_lab = next(self.trainIter)
            except StopIteration:
                self.trainIter = iter(self.train_loader)
                data_lab = next(self.trainIter)
            try:
                data_unlab = next(self.trainIter_unlab)
            except StopIteration:
                self.trainIter_unlab = iter(self.train_loader_unlab)
                data_unlab = next(self.trainIter_unlab)

            # get the inputs
            x0   = self.convert_tensor_type(data_lab['image'])
            y0   = self.convert_tensor_type(data_lab['label_prob'])  
            x1   = self.convert_tensor_type(data_unlab['image'])
            inputs = torch.cat([x0, x1], dim = 0)               
            inputs, y0 = inputs.to(self.device), y0.to(self.device)
            inputs_ema = x1 + torch.clamp(torch.randn_like(x1) * 0.1, -0.2, 0.2)
            inputs_ema = inputs_ema.to(self.device)

            # zero the parameter gradients
            self.optimizer.zero_grad()
                
            outputs = self.net(inputs)
            n0 = list(x0.shape)[0] 
            p0, p1  = torch.tensor_split(outputs, [n0,], dim = 0)
            outputs_soft = torch.softmax(outputs, dim=1)
            p0_soft, p1_soft = torch.tensor_split(outputs_soft, [n0,], dim = 0)
            loss_sup = self.get_loss_value(data_lab, p0, y0)

            with torch.no_grad():
                outputs_ema = self.net_ema(inputs_ema)
                p1_ema_soft = torch.softmax(outputs_ema, dim=1)
            square_error = torch.square(p1_soft - p1_ema_soft)

            # the forward pass number for uncertainty estimation
            T = ssl_cfg.get("uamt_mcdroput_n", 8)
            preds = torch.zeros([T] + list(p1.shape)).to(self.device)
            for i in range(T//2):
                ema_inputs_r = torch.cat([x1, x1], dim = 0)
                ema_inputs_r = ema_inputs_r + \
                    torch.clamp(torch.randn_like(ema_inputs_r) * 0.1, -0.2, 0.2)
                ema_inputs_r = ema_inputs_r.to(self.device)
                with torch.no_grad():
                    ema_outputs_r = self.net_ema(ema_inputs_r)
                # reshape from [2B, C, D, H, W] to [2, B, C, D, H, W]
                preds[2*i:2*(i+1)] = ema_outputs_r.reshape([2]+list(p1.shape))
            preds = torch.softmax(preds, dim = 2)
            preds = torch.mean(preds, dim = 0)
            uncertainty = -1.0 * torch.sum(preds*torch.log(preds + 1e-6),
                 dim=1, keepdim=True)
            
            rampup_ratio = get_rampup_ratio(self.glob_it, rampup_start, rampup_end, "sigmoid")
            class_num = list(y0.shape)[1]
            threshold = (0.75+0.25*rampup_ratio)*np.log(class_num)
            mask      = (uncertainty < threshold).float()
            loss_reg  = torch.sum(mask*square_error)/(2*torch.sum(mask)+1e-16)

            regular_w = ssl_cfg.get('regularize_w', 0.1) * rampup_ratio
            loss = loss_sup + regular_w*loss_reg

            loss.backward()
            self.optimizer.step()
            if(self.scheduler is not None and \
                not isinstance(self.scheduler, lr_scheduler.ReduceLROnPlateau)):
                self.scheduler.step()


            # update EMA
            alpha = ssl_cfg.get('ema_decay', 0.99)
            alpha = min(1 - 1 / (iter_max + 1), alpha)
            for ema_param, param in zip(self.net_ema.parameters(), self.net.parameters()):
                ema_param.data.mul_(alpha).add_(1 - alpha, param.data)

            train_loss = train_loss + loss.item()
            train_loss_sup = train_loss_sup + loss_sup.item()
            train_loss_reg = train_loss_reg + loss_reg.item() 
            # get dice evaluation for each class in annotated images
            if(isinstance(p0, tuple) or isinstance(p0, list)):
                p0 = p0[0] 
            p0_argmax = torch.argmax(p0, dim = 1, keepdim = True)
            p0_soft   = get_soft_label(p0_argmax, class_num, self.tensor_type)
            p0_soft, y0 = reshape_prediction_and_ground_truth(p0_soft, y0) 
            dice_list   = get_classwise_dice(p0_soft, y0)
            train_dice_list.append(dice_list.cpu().numpy())
        train_avg_loss = train_loss / iter_valid
        train_avg_loss_sup = train_loss_sup / iter_valid
        train_avg_loss_reg = train_loss_reg / iter_valid
        train_cls_dice = np.asarray(train_dice_list).mean(axis = 0)
        train_avg_dice = train_cls_dice.mean()

        train_scalers = {'loss': train_avg_loss, 'loss_sup':train_avg_loss_sup,
            'loss_reg':train_avg_loss_reg, 'regular_w':regular_w,
            'avg_dice':train_avg_dice,     'class_dice': train_cls_dice}
        return train_scalers


# -*- coding: utf-8 -*-
from __future__ import print_function, division
import logging
import numpy as np
import random
import torch
import torchvision.transforms as transforms
from pymic.io.nifty_dataset import NiftyDataset
from pymic.loss.seg.util import get_soft_label
from pymic.loss.seg.util import reshape_prediction_and_ground_truth
from pymic.loss.seg.util import get_classwise_dice
from pymic.loss.seg.ssl import EntropyLoss
from pymic.net_run.agent_seg import SegmentationAgent
from pymic.transform.trans_dict import TransformDict

class SSLSegAgent(SegmentationAgent):
    """
    Implementation of the following paper:
    Yves Grandvalet and Yoshua Bengio, 
    Semi-supervised Learningby Entropy Minimization.
    NeurIPS, 2005. 
    """
    def __init__(self, config, stage = 'train'):
        super(SSLSegAgent, self).__init__(config, stage)
        self.transform_dict  = TransformDict
        self.train_set_unlab = None 

    def get_unlabeled_dataset_from_config(self):
        root_dir  = self.config['dataset']['root_dir']
        modal_num = self.config['dataset'].get('modal_num', 1)
        transform_names = self.config['dataset']['train_transform_unlab']
        
        self.transform_list  = []
        if(transform_names is None or len(transform_names) == 0):
            data_transform = None 
        else:
            transform_param = self.config['dataset']
            transform_param['task'] = 'segmentation' 
            for name in transform_names:
                if(name not in self.transform_dict):
                    raise(ValueError("Undefined transform {0:}".format(name))) 
                one_transform = self.transform_dict[name](transform_param)
                self.transform_list.append(one_transform)
            data_transform = transforms.Compose(self.transform_list)

        csv_file = self.config['dataset'].get('train_csv_unlab', None)
        dataset  = NiftyDataset(root_dir=root_dir,
                                csv_file  = csv_file,
                                modal_num = modal_num,
                                with_label= False,
                                transform = data_transform )
        return dataset

    def create_dataset(self):
        super(SSLSegAgent, self).create_dataset()
        if(self.stage == 'train'):
            if(self.train_set_unlab is None):
                self.train_set_unlab = self.get_unlabeled_dataset_from_config()
            if(self.deterministic):
                def worker_init_fn(worker_id):
                    random.seed(self.random_seed+worker_id)
                worker_init = worker_init_fn
            else:
                worker_init = None

            bn_train_unlab = self.config['dataset']['train_batch_size_unlab']
            num_worker = self.config['dataset'].get('num_workder', 16)
            self.train_loader_unlab = torch.utils.data.DataLoader(self.train_set_unlab, 
                batch_size = bn_train_unlab, shuffle=True, num_workers= num_worker,
                worker_init_fn=worker_init)

    def write_scalars(self, train_scalars, valid_scalars, lr_value, glob_it):
        loss_scalar ={'train':train_scalars['loss'], 
                      'valid':valid_scalars['loss']}
        loss_sup_scalar  = {'train':train_scalars['loss_sup']}
        loss_upsup_scalar  = {'train':train_scalars['loss_reg']}
        dice_scalar ={'train':train_scalars['avg_dice'], 'valid':valid_scalars['avg_dice']}
        self.summ_writer.add_scalars('loss', loss_scalar, glob_it)
        self.summ_writer.add_scalars('loss_sup', loss_sup_scalar, glob_it)
        self.summ_writer.add_scalars('loss_reg', loss_upsup_scalar, glob_it)
        self.summ_writer.add_scalars('regular_w', {'regular_w':train_scalars['regular_w']}, glob_it)
        self.summ_writer.add_scalars('lr', {"lr": lr_value}, glob_it)
        self.summ_writer.add_scalars('dice', dice_scalar, glob_it)
        class_num = self.config['network']['class_num']
        for c in range(class_num):
            cls_dice_scalar = {'train':train_scalars['class_dice'][c], \
                'valid':valid_scalars['class_dice'][c]}
            self.summ_writer.add_scalars('class_{0:}_dice'.format(c), cls_dice_scalar, glob_it)
        logging.info('train loss {0:.4f}, avg dice {1:.4f} '.format(
            train_scalars['loss'], train_scalars['avg_dice']) + "[" + \
            ' '.join("{0:.4f}".format(x) for x in train_scalars['class_dice']) + "]")        
        logging.info('valid loss {0:.4f}, avg dice {1:.4f} '.format(
            valid_scalars['loss'], valid_scalars['avg_dice']) + "[" + \
            ' '.join("{0:.4f}".format(x) for x in valid_scalars['class_dice']) + "]")  

    def train_valid(self):
        self.trainIter_unlab = iter(self.train_loader_unlab)   
        super(SSLSegAgent, self).train_valid()    



# -*- coding: utf-8 -*-
from __future__ import print_function, division
from pymic.transform.intensity import  *
from pymic.transform.gray2rgb import GrayscaleToRGB
from pymic.transform.flip import RandomFlip
from pymic.transform.intensity import GaussianNoise
from pymic.transform.pad import Pad
from pymic.transform.rotate import RandomRotate
from pymic.transform.rescale import Rescale, RandomRescale  
from pymic.transform.threshold import * 
from pymic.transform.normalize import *
from pymic.transform.crop import *
from pymic.transform.label_convert import * 

TransformDict = {
    'ChannelWiseThreshold': ChannelWiseThreshold,
    'ChannelWiseThresholdWithNormalize': ChannelWiseThresholdWithNormalize,
    'CropWithBoundingBox': CropWithBoundingBox,
    'CenterCrop': CenterCrop,
    'GrayscaleToRGB': GrayscaleToRGB,
    'GammaCorrection': GammaCorrection,
    'GaussianNoise': GaussianNoise,
    'LabelConvert': LabelConvert,
    'LabelConvertNonzero': LabelConvertNonzero,
    'LabelToProbability': LabelToProbability,
    'NormalizeWithMeanStd': NormalizeWithMeanStd,
    'NormalizeWithMinMax': NormalizeWithMinMax,
    'NormalizeWithPercentiles': NormalizeWithPercentiles,
    'PartialLabelToProbability':PartialLabelToProbability,
    'RandomCrop': RandomCrop,
    'RandomResizedCrop': RandomResizedCrop,
    'RandomRescale': RandomRescale,
    'RandomFlip': RandomFlip,
    'RandomRotate': RandomRotate,
    'ReduceLabelDim': ReduceLabelDim,
    'Rescale': Rescale,
    'Pad': Pad,
}



# -*- coding: utf-8 -*-
"""
An modification the U-Net to obtain multi-scale prediction according to 
the URPC paper (MICCAI 2021):
    Xiangde Luo, Wenjun Liao, Jienneg Chen, Tao Song, Yinan Chen, 
    Shichuan Zhang, Nianyong Chen, Guotai Wang, Shaoting Zhang:
    Efficient Semi-Supervised Gross Target Volume of Nasopharyngeal Carcinoma 
    Segmentation via Uncertainty Rectified Pyramid Consistency. 
    MICCAI 2021: 318-329
    https://link.springer.com/chapter/10.1007/978-3-030-87196-3_30 
Also see: https://github.com/HiLab-git/SSL4MIS/blob/master/code/networks/unet.py
"""
from __future__ import print_function, division

import torch
import torch.nn as nn
import numpy as np 
from torch.distributions.uniform import Uniform
from pymic.net.net2d.unet2d import ConvBlock, DownBlock, UpBlock

def FeatureDropout(x):
    attention = torch.mean(x, dim=1, keepdim=True)
    max_val, _ = torch.max(attention.view(
        x.size(0), -1), dim=1, keepdim=True)
    threshold = max_val * np.random.uniform(0.7, 0.9)
    threshold = threshold.view(x.size(0), 1, 1, 1).expand_as(attention)
    drop_mask = (attention < threshold).float()
    x = x.mul(drop_mask)
    return x

class FeatureNoise(nn.Module):
    def __init__(self, uniform_range=0.3):
        super(FeatureNoise, self).__init__()
        self.uni_dist = Uniform(-uniform_range, uniform_range)

    def feature_based_noise(self, x):
        noise_vector = self.uni_dist.sample(
            x.shape[1:]).to(x.device).unsqueeze(0)
        x_noise = x.mul(noise_vector) + x
        return x_noise

    def forward(self, x):
        x = self.feature_based_noise(x)
        return x

class UNet2D_URPC(nn.Module):
    def __init__(self, params):
        super(UNet2D_URPC, self).__init__()
        self.params    = params
        self.in_chns   = self.params['in_chns']
        self.ft_chns   = self.params['feature_chns']
        self.n_class   = self.params['class_num']
        self.bilinear  = self.params['bilinear']
        self.dropout   = self.params['dropout']
        assert(len(self.ft_chns) == 5)

        self.in_conv= ConvBlock(self.in_chns, self.ft_chns[0], self.dropout[0])
        self.down1  = DownBlock(self.ft_chns[0], self.ft_chns[1], self.dropout[1])
        self.down2  = DownBlock(self.ft_chns[1], self.ft_chns[2], self.dropout[2])
        self.down3  = DownBlock(self.ft_chns[2], self.ft_chns[3], self.dropout[3])
        self.down4  = DownBlock(self.ft_chns[3], self.ft_chns[4], self.dropout[4])
        self.up1 = UpBlock(self.ft_chns[4], self.ft_chns[3], self.ft_chns[3], 0.0, self.bilinear) 
        self.up2 = UpBlock(self.ft_chns[3], self.ft_chns[2], self.ft_chns[2], 0.0, self.bilinear) 
        self.up3 = UpBlock(self.ft_chns[2], self.ft_chns[1], self.ft_chns[1], 0.0, self.bilinear) 
        self.up4 = UpBlock(self.ft_chns[1], self.ft_chns[0], self.ft_chns[0], 0.0, self.bilinear) 
    
        self.out_conv = nn.Conv2d(self.ft_chns[0], self.n_class,  
                                  kernel_size = 3, padding = 1)
        self.out_conv_dp4 = nn.Conv2d(self.ft_chns[4], self.n_class,
                                      kernel_size=3, padding=1)
        self.out_conv_dp3 = nn.Conv2d(self.ft_chns[3], self.n_class,
                                      kernel_size=3, padding=1)
        self.out_conv_dp2 = nn.Conv2d(self.ft_chns[2], self.n_class,
                                      kernel_size=3, padding=1)
        self.out_conv_dp1 = nn.Conv2d(self.ft_chns[1], self.n_class,
                                      kernel_size=3, padding=1)
        self.feature_noise = FeatureNoise()

    def forward(self, x):
        x_shape = list(x.shape)
        if(len(x_shape) == 5):
          [N, C, D, H, W] = x_shape
          new_shape = [N*D, C, H, W]
          x = torch.transpose(x, 1, 2)
          x = torch.reshape(x, new_shape)
        x0 = self.in_conv(x)
        x1 = self.down1(x0)
        x2 = self.down2(x1)
        x3 = self.down3(x2)
        x4 = self.down4(x3)
        
        x = self.up1(x4, x3)
        if self.training:
            x = nn.functional.dropout(x, p=0.5)
        dp3_out = self.out_conv_dp3(x)

        x = self.up2(x, x2)
        if self.training:
            x = FeatureDropout(x)
        dp2_out = self.out_conv_dp2(x)

        x = self.up3(x, x1)
        if self.training:
            x = self.feature_noise(x)
        dp1_out = self.out_conv_dp1(x)

        x = self.up4(x, x0)
        dp0_out = self.out_conv(x)

        out_shape = list(dp0_out.shape)[2:]
        dp3_out = nn.functional.interpolate(dp3_out, out_shape)
        dp2_out = nn.functional.interpolate(dp2_out, out_shape)
        dp1_out = nn.functional.interpolate(dp1_out, out_shape)
        out = [dp0_out, dp1_out, dp2_out, dp3_out]

        if(len(x_shape) == 5):
            new_shape = [N, D] + list(dp0_out.shape)[1:]
            for i in range(len(out)):
                out[i] = torch.transpose(torch.reshape(out[i], new_shape), 1, 2)
        return out


# -*- coding: utf-8 -*-
"""
Extention of U-Net with two decoders. The network was introduced in
the following paper:
    Xiangde Luo, Minhao Hu, Wenjun Liao, Shuwei Zhai, Tao Song, Guotai Wang,
    Shaoting Zhang. ScribblScribble-Supervised Medical Image Segmentation via 
    Dual-Branch Network and Dynamically Mixed Pseudo Labels Supervision.
    MICCAI 2022. 
"""
from __future__ import print_function, division

import torch
import torch.nn as nn
from pymic.net.net2d.unet2d import *

class UNet2D_DualBranch(nn.Module):
    def __init__(self, params):
        super(UNet2D_DualBranch, self).__init__()
        self.output_mode = params.get("output_mode", "average")
        self.encoder  = Encoder(params)
        self.decoder1 = Decoder(params)    
        self.decoder2 = Decoder(params)        

    def forward(self, x):
        x_shape = list(x.shape)
        if(len(x_shape) == 5):
          [N, C, D, H, W] = x_shape
          new_shape = [N*D, C, H, W]
          x = torch.transpose(x, 1, 2)
          x = torch.reshape(x, new_shape)

        f = self.encoder(x)
        output1 = self.decoder1(f)
        output2 = self.decoder2(f)
        if(len(x_shape) == 5):
            new_shape = [N, D] + list(output1.shape)[1:]
            output1 = torch.reshape(output1, new_shape)
            output1 = torch.transpose(output1, 1, 2)
            output2 = torch.reshape(output2, new_shape)
            output2 = torch.transpose(output2, 1, 2)

        if(self.training):
          return output1, output2
        else:
          if(self.output_mode == "average"):
            return (output1 + output2)/2
          elif(self.output_mode == "first"):
            return output1
          else:
            return output2



# -*- coding: utf-8 -*-
from __future__ import print_function, division

import os
import random
import logging
import torch
import numpy as np
import torch.optim as optim
from abc import ABCMeta, abstractmethod
from pymic.net_run.get_optimizer import get_lr_scheduler, get_optimizer

def seed_torch(seed=1):
    random.seed(seed)
    os.environ['PYTHONHASHSEED'] = str(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed) # if using multi-GPU.
    torch.backends.cudnn.benchmark = False
    torch.backends.cudnn.deterministic = True

class NetRunAgent(object):
    __metaclass__ = ABCMeta
    def __init__(self, config, stage = 'train'):
        assert(stage in ['train', 'inference', 'test'])
        self.config = config
        self.stage  = stage
        if(stage == 'inference'):
            self.stage = 'test'
        self.train_set = None 
        self.valid_set = None 
        self.test_set  = None
        self.net       = None
        self.optimizer = None
        self.scheduler = None 
        self.loss_dict = None 
        self.transform_dict  = None
        self.inferer   = None
        self.tensor_type   = config['dataset']['tensor_type']
        self.task_type     = config['dataset']['task_type'] #cls, cls_mtbc, seg
        self.deterministic = config['training'].get('deterministic', True)
        self.random_seed   = config['training'].get('random_seed', 1)
        if(self.deterministic):
            seed_torch(self.random_seed)
            logging.info("deterministric is true")
        
    def set_datasets(self, train_set, valid_set, test_set):
        self.train_set = train_set
        self.valid_set = valid_set
        self.test_set  = test_set

    def set_transform_dict(self, custom_transform_dict):
        self.transform_dict = custom_transform_dict

    def set_network(self, net):
        self.net = net 

    def set_loss_dict(self, loss_dict):
        self.loss_dict = loss_dict

    def set_optimizer(self, optimizer):
        self.optimizer = optimizer
    
    def set_scheduler(self, scheduler):
        self.scheduler = scheduler
    
    def set_inferer(self, inferer):
        self.inferer = inferer

    def get_checkpoint_name(self):
        ckpt_mode = self.config['testing']['ckpt_mode']
        if(ckpt_mode == 0 or ckpt_mode == 1):
            ckpt_dir    = self.config['training']['ckpt_save_dir']
            ckpt_prefix = self.config['training'].get('ckpt_prefix', None)
            if(ckpt_prefix is None):
                ckpt_prefix = ckpt_dir.split('/')[-1]
            txt_name = ckpt_dir + '/' + ckpt_prefix
            txt_name += "_latest.txt" if ckpt_mode == 0 else "_best.txt"
            with open(txt_name, 'r') as txt_file:
                it_num = txt_file.read().replace('\n', '') 
                ckpt_name = "{0:}/{1:}_{2:}.pt".format(ckpt_dir, ckpt_prefix, it_num)
        else:
            ckpt_name =  self.config['testing']['ckpt_name']
        return ckpt_name

    @abstractmethod    
    def get_stage_dataset_from_config(self, stage):
        raise(ValueError("not implemented"))

    @abstractmethod
    def get_parameters_to_update(self):
        raise(ValueError("not implemented"))

    @abstractmethod
    def create_network(self):
        raise(ValueError("not implemented"))

    @abstractmethod
    def training(self):
        raise(ValueError("not implemented"))
        
    @abstractmethod
    def validation(self):
        raise(ValueError("not implemented"))

    @abstractmethod
    def train_valid(self):
        raise(ValueError("not implemented"))
    
    @abstractmethod
    def infer(self):
        raise(ValueError("not implemented"))

    def create_dataset(self):
        if(self.stage == 'train'):
            if(self.train_set is None):
                self.train_set = self.get_stage_dataset_from_config('train')
            if(self.valid_set is None):
                self.valid_set = self.get_stage_dataset_from_config('valid')
            if(self.deterministic):
                def worker_init_fn(worker_id):
                    # workder_seed = self.random_seed+worker_id 
                    workder_seed = torch.initial_seed() % 2 ** 32
                    np.random.seed(workder_seed)
                    random.seed(workder_seed)                    
                worker_init = worker_init_fn
            else:
                worker_init = None

            bn_train = self.config['dataset']['train_batch_size']
            bn_valid = self.config['dataset'].get('valid_batch_size', 1)
            num_worker = self.config['dataset'].get('num_workder', 16)
            g_train, g_valid = torch.Generator(), torch.Generator()
            g_train.manual_seed(self.random_seed)
            g_valid.manual_seed(self.random_seed)
            self.train_loader = torch.utils.data.DataLoader(self.train_set, 
                batch_size = bn_train, shuffle=True, num_workers= num_worker,
                worker_init_fn=worker_init, generator = g_train)
            self.valid_loader = torch.utils.data.DataLoader(self.valid_set, 
                batch_size = bn_valid, shuffle=False, num_workers= num_worker,
                worker_init_fn=worker_init, generator = g_valid)
        else:
            bn_test = self.config['dataset'].get('test_batch_size', 1)
            if(self.test_set  is None):
                self.test_set  = self.get_stage_dataset_from_config('test')
            self.test_loader = torch.utils.data.DataLoader(self.test_set, 
                batch_size = bn_test, shuffle=False, num_workers= bn_test)
       
    def create_optimizer(self, params):
        opt_params = self.config['training']
        if(self.optimizer is None):
            self.optimizer = get_optimizer(opt_params['optimizer'],
                    params, opt_params)
        last_iter = -1
        if(self.checkpoint is not None):
            self.optimizer.load_state_dict(self.checkpoint['optimizer_state_dict'])
            last_iter = self.checkpoint['iteration'] - 1
        if(self.scheduler is None):
            opt_params["last_iter"] = last_iter
            self.scheduler = get_lr_scheduler(self.optimizer, opt_params)

    def convert_tensor_type(self, input_tensor):
        if(self.tensor_type == 'float'):
            return input_tensor.float()
        else:
            return input_tensor.double()

    def run(self):
        self.create_dataset()
        self.create_network()
        if(self.stage == 'train'):
            self.train_valid()
        else:
            self.infer()




# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function
import csv
import os
import sys
import math
import pandas as pd
import random
import GeodisTK
import configparser
import numpy as np
from scipy import ndimage
from pymic.io.image_read_write import *
from pymic.util.image_process import *
from pymic.util.parse_config import parse_config

# Dice evaluation
def binary_dice(s, g, resize = False):
    """
    calculate the Dice score of two N-d volumes.
    s: the segmentation volume of numpy array
    g: the ground truth volume of numpy array
    resize: if s and g have different shapes, resize s to match g.
    """
    assert(len(s.shape)== len(g.shape))
    if(resize):
        size_match = True
        for i in range(len(s.shape)):
            if(s.shape[i] != g.shape[i]):
                size_match = False
                break
        if(size_match is False):
            s = resize_ND_volume_to_given_shape(s, g.shape, order = 0)
    prod = np.multiply(s, g)
    s0 = prod.sum()
    s1 = s.sum()
    s2 = g.sum()
    dice = (2.0*s0 + 1e-5)/(s1 + s2 + 1e-5)
    return dice

def dice_of_images(s_name, g_name):
    s = load_image_as_nd_array(s_name)['data_array']
    g = load_image_as_nd_array(g_name)['data_array']
    dice = binary_dice(s, g)
    return dice

# IOU evaluation
def binary_iou(s,g):
    assert(len(s.shape)== len(g.shape))
    intersecion = np.multiply(s, g)
    union = np.asarray(s + g >0, np.float32)
    iou = (intersecion.sum() + 1e-5)/(union.sum() + 1e-5)
    return iou

def iou_of_images(s_name, g_name):
    s = load_image_as_nd_array(s_name)['data_array']
    g = load_image_as_nd_array(g_name)['data_array']
    margin = (3, 8, 8)
    g = get_detection_binary_bounding_box(g, margin)
    return binary_iou(s, g)

# Hausdorff and ASSD evaluation
def get_edge_points(img):
    """
    get edge points of a binary segmentation result
    """
    dim = len(img.shape)
    if(dim == 2):
        strt = ndimage.generate_binary_structure(2,1)
    else:
        strt = ndimage.generate_binary_structure(3,1)
    ero  = ndimage.morphology.binary_erosion(img, strt)
    edge = np.asarray(img, np.uint8) - np.asarray(ero, np.uint8) 
    return edge 


def binary_hd95(s, g, spacing = None):
    """
    get the hausdorff distance between a binary segmentation and the ground truth
    inputs:
        s: a 3D or 2D binary image for segmentation
        g: a 2D or 2D binary image for ground truth
        spacing: a list for image spacing, length should be 3 or 2
    """
    s_edge = get_edge_points(s)
    g_edge = get_edge_points(g)
    image_dim = len(s.shape)
    assert(image_dim == len(g.shape))
    if(spacing == None):
        spacing = [1.0] * image_dim
    else:
        assert(image_dim == len(spacing))
    img = np.zeros_like(s)
    if(image_dim == 2):
        s_dis = GeodisTK.geodesic2d_raster_scan(img, s_edge, 0.0, 2)
        g_dis = GeodisTK.geodesic2d_raster_scan(img, g_edge, 0.0, 2)
    elif(image_dim ==3):
        s_dis = GeodisTK.geodesic3d_raster_scan(img, s_edge, spacing, 0.0, 2)
        g_dis = GeodisTK.geodesic3d_raster_scan(img, g_edge, spacing, 0.0, 2)

    dist_list1 = s_dis[g_edge > 0]
    dist_list1 = sorted(dist_list1)
    dist1 = dist_list1[int(len(dist_list1)*0.95)]
    dist_list2 = g_dis[s_edge > 0]
    dist_list2 = sorted(dist_list2)
    dist2 = dist_list2[int(len(dist_list2)*0.95)]
    return max(dist1, dist2)


def binary_assd(s, g, spacing = None):
    """
    get the average symetric surface distance between a binary segmentation and the ground truth
    inputs:
        s: a 3D or 2D binary image for segmentation
        g: a 2D or 2D binary image for ground truth
        spacing: a list for image spacing, length should be 3 or 2
    """
    s_edge = get_edge_points(s)
    g_edge = get_edge_points(g)
    image_dim = len(s.shape)
    assert(image_dim == len(g.shape))
    if(spacing == None):
        spacing = [1.0] * image_dim
    else:
        assert(image_dim == len(spacing))
    img = np.zeros_like(s)
    if(image_dim == 2):
        s_dis = GeodisTK.geodesic2d_raster_scan(img, s_edge, 0.0, 2)
        g_dis = GeodisTK.geodesic2d_raster_scan(img, g_edge, 0.0, 2)
    elif(image_dim ==3):
        s_dis = GeodisTK.geodesic3d_raster_scan(img, s_edge, spacing, 0.0, 2)
        g_dis = GeodisTK.geodesic3d_raster_scan(img, g_edge, spacing, 0.0, 2)

    ns = s_edge.sum()
    ng = g_edge.sum()
    s_dis_g_edge = s_dis * g_edge
    g_dis_s_edge = g_dis * s_edge
    assd = (s_dis_g_edge.sum() + g_dis_s_edge.sum()) / (ns + ng) 
    return assd

# relative volume error evaluation
def binary_relative_volume_error(s_volume, g_volume):
    s_v = float(s_volume.sum())
    g_v = float(g_volume.sum())
    assert(g_v > 0)
    rve = abs(s_v - g_v)/g_v
    return rve

def get_binary_evaluation_score(s_volume, g_volume, spacing, metric):
    if(len(s_volume.shape) == 4):
        assert(s_volume.shape[0] == 1 and g_volume.shape[0] == 1)
        s_volume = np.reshape(s_volume, s_volume.shape[1:])
        g_volume = np.reshape(g_volume, g_volume.shape[1:])
    if(s_volume.shape[0] == 1):
        s_volume = np.reshape(s_volume, s_volume.shape[1:])
        g_volume = np.reshape(g_volume, g_volume.shape[1:])
    metric_lower = metric.lower()

    if(metric_lower == "dice"):
        score = binary_dice(s_volume, g_volume)

    elif(metric_lower == "iou"):
        score = binary_iou(s_volume,g_volume)

    elif(metric_lower == 'assd'):
        score = binary_assd(s_volume, g_volume, spacing)

    elif(metric_lower == "hd95"):
        score = binary_hd95(s_volume, g_volume, spacing)

    elif(metric_lower == "rve"):
        score = binary_relative_volume_error(s_volume, g_volume)

    elif(metric_lower == "volume"):
        voxel_size = 1.0
        for dim in range(len(spacing)):
            voxel_size = voxel_size * spacing[dim]
        score = g_volume.sum()*voxel_size
    else:
        raise ValueError("unsupported evaluation metric: {0:}".format(metric))

    return score

def get_multi_class_evaluation_score(s_volume, g_volume, label_list, fuse_label, spacing, metric):
    if(fuse_label):
        s_volume_sub = np.zeros_like(s_volume)
        g_volume_sub = np.zeros_like(g_volume)
        for lab in label_list:
            s_volume_sub = s_volume_sub + np.asarray(s_volume == lab, np.uint8)
            g_volume_sub = g_volume_sub + np.asarray(g_volume == lab, np.uint8)
        label_list = [1]
        s_volume = np.asarray(s_volume_sub > 0, np.uint8)
        g_volume = np.asarray(g_volume_sub > 0, np.uint8)
    score_list = []
    for label in label_list:
        temp_score = get_binary_evaluation_score(s_volume == label, g_volume == label,
                    spacing, metric)
        score_list.append(temp_score)
    return score_list

def evaluation(config_file):
    config = parse_config(config_file)['evaluation']
    metric = config['metric']
    label_list = config['label_list']
    label_fuse = config.get('label_fuse', False)
    organ_name = config['organ_name']
    gt_root    = config['ground_truth_folder_root']
    seg_root   = config['segmentation_folder_root']
    if(not(isinstance(seg_root, tuple) or isinstance(seg_root, list))):
        seg_root = [seg_root]
    image_pair_csv = config['evaluation_image_pair']
    ground_truth_label_convert_source = config.get('ground_truth_label_convert_source', None)
    ground_truth_label_convert_target = config.get('ground_truth_label_convert_target', None)
    segmentation_label_convert_source = config.get('segmentation_label_convert_source', None)
    segmentation_label_convert_target = config.get('segmentation_label_convert_target', None)

    image_items = pd.read_csv(image_pair_csv)
    item_num = len(image_items)
    for seg_root_n in seg_root:
        score_all_data = []
        name_score_list= []
        for i in range(item_num):
            gt_name  = image_items.iloc[i, 0]
            seg_name = image_items.iloc[i, 1]
            # seg_name = seg_name.replace(".nii.gz", "_pred.nii.gz")
            gt_full_name  = gt_root  + '/' + gt_name
            seg_full_name = seg_root_n + '/' + seg_name
            
            s_dict = load_image_as_nd_array(seg_full_name)
            g_dict = load_image_as_nd_array(gt_full_name)
            s_volume = s_dict["data_array"]; s_spacing = s_dict["spacing"]
            g_volume = g_dict["data_array"]; g_spacing = g_dict["spacing"]
            # for dim in range(len(s_spacing)):
            #     assert(s_spacing[dim] == g_spacing[dim])
            if((ground_truth_label_convert_source is not None) and \
                ground_truth_label_convert_target is not None):
                g_volume = convert_label(g_volume, ground_truth_label_convert_source, \
                    ground_truth_label_convert_target)

            if((segmentation_label_convert_source is not None) and \
                segmentation_label_convert_target is not None):
                s_volume = convert_label(s_volume, segmentation_label_convert_source, \
                    segmentation_label_convert_target)

            score_vector = get_multi_class_evaluation_score(s_volume, g_volume, label_list, 
                label_fuse, s_spacing, metric )
            if(len(label_list) > 1):
                score_vector.append(np.asarray(score_vector).mean())
            score_all_data.append(score_vector)
            name_score_list.append([seg_name] + score_vector)
            print(seg_name, score_vector)
        score_all_data = np.asarray(score_all_data)
        score_mean = score_all_data.mean(axis = 0)
        score_std  = score_all_data.std(axis = 0)
        name_score_list.append(['mean'] + list(score_mean))
        name_score_list.append(['std'] + list(score_std))
    
        # save the result as csv 
        score_csv = "{0:}/{1:}_{2:}_all.csv".format(seg_root_n, organ_name, metric)
        with open(score_csv, mode='w') as csv_file:
            csv_writer = csv.writer(csv_file, delimiter=',', 
                            quotechar='"',quoting=csv.QUOTE_MINIMAL)
            head = ['image'] + ["class_{0:}".format(i) for i in label_list]
            if(len(label_list) > 1):
                head = head + ["average"]
            csv_writer.writerow(head)
            for item in name_score_list:
                csv_writer.writerow(item)

        print("{0:} mean ".format(metric), score_mean)
        print("{0:} std  ".format(metric), score_std) 

def main():
    if(len(sys.argv) < 2):
        print('Number of arguments should be 2. e.g.')
        print('    pymic_evaluate_seg config.cfg')
        exit()
    config_file = str(sys.argv[1])
    assert(os.path.isfile(config_file))
    evaluation(config_file)
    
if __name__ == '__main__':
    main()




# -*- coding: utf-8 -*-
from __future__ import print_function, division
import logging 
import os
import sys
from pymic.util.parse_config import *
from pymic.net_run_wsl.wsl_em import WSLEntropyMinimization
from pymic.net_run_wsl.wsl_gatedcrf import WSLGatedCRF
from pymic.net_run_wsl.wsl_mumford_shah import WSLMumfordShah
from pymic.net_run_wsl.wsl_tv import WSLTotalVariation
from pymic.net_run_wsl.wsl_ustm import WSLUSTM
from pymic.net_run_wsl.wsl_dmpls import WSLDMPLS

WSLMethodDict = {'EntropyMinimization': WSLEntropyMinimization,
    'GatedCRF': WSLGatedCRF,
    'MumfordShah': WSLMumfordShah,
    'TotalVariation': WSLTotalVariation,
    'USTM': WSLUSTM,
    'DMPLS': WSLDMPLS}

def main():
    if(len(sys.argv) < 3):
        print('Number of arguments should be 3. e.g.')
        print('   pymic_ssl train config.cfg')
        exit()
    stage    = str(sys.argv[1])
    cfg_file = str(sys.argv[2])
    config   = parse_config(cfg_file)
    config   = synchronize_config(config)
    log_dir  = config['training']['ckpt_save_dir']
    if(not os.path.exists(log_dir)):
        os.mkdir(log_dir)
    logging.basicConfig(filename=log_dir+"/log_{0:}.txt".format(stage), level=logging.INFO,
                        format='%(message)s')
    logging.getLogger().addHandler(logging.StreamHandler(sys.stdout))
    logging_config(config)
    wsl_method = config['weakly_supervised_learning']['wsl_method']
    agent = WSLMethodDict[wsl_method](config, stage)
    agent.run()

if __name__ == "__main__":
    main()

    


# -*- coding: utf-8 -*-
from __future__ import print_function, division

import torch
import json
import math
import random
import numpy as np
from scipy import ndimage
from pymic.transform.abstract_transform import AbstractTransform
from pymic.util.image_process import *


class Rescale(AbstractTransform):
    """Rescale the image in a sample to a given size."""
    def __init__(self, params):
        """
        output_size (tuple/list or int): Desired output size. 
            If tuple/list, output_size should in the format of [D, H, W] or [H, W].
            Channel number is kept the same as the input. If D is None, the input image
            is only reslcaled in 2D.
            If int, the smallest axis is matched to output_size keeping 
            aspect ratio the same.
        """
        super(Rescale, self).__init__(params)
        self.output_size = params["Rescale_output_size".lower()]
        self.inverse     = params.get("Rescale_inverse".lower(), True)
        assert isinstance(self.output_size, (int, list, tuple))

    def __call__(self, sample):
        image = sample['image']
        input_shape = image.shape
        input_dim   = len(input_shape) - 1
        
        if isinstance(self.output_size, (list, tuple)):
            output_size = self.output_size
            if(output_size[0] is None):
                output_size[0] = input_shape[1]
            assert(len(output_size) == input_dim)
        else:
            min_edge = min(input_shape[1:])
            output_size = [self.output_size * input_shape[i+1] / min_edge for \
                            i in range(input_dim)]
        scale = [(output_size[i] + 0.0)/input_shape[1:][i] for i in range(input_dim)]
        scale = [1.0] + scale
        image_t = ndimage.interpolation.zoom(image, scale, order = 1)

        sample['image'] = image_t
        sample['Rescale_origin_shape'] = json.dumps(input_shape)
        if('label' in sample and self.task == 'segmentation'):
            label = sample['label']
            label = ndimage.interpolation.zoom(label, scale, order = 0)
            sample['label'] = label
        if('pixel_weight' in sample and self.task == 'segmentation'):
            weight = sample['pixel_weight']
            weight = ndimage.interpolation.zoom(weight, scale, order = 1)
            sample['pixel_weight'] = weight
        
        return sample

    def inverse_transform_for_prediction(self, sample):
        ''' rescale sample['predict'] (5D or 4D) to the original spatial shape.
        origin_shape is a 4D or 3D vector as saved in __call__().'''
        if(isinstance(sample['Rescale_origin_shape'], list) or \
            isinstance(sample['Rescale_origin_shape'], tuple)):
            origin_shape = json.loads(sample['Rescale_origin_shape'][0])
        else:
            origin_shape = json.loads(sample['Rescale_origin_shape'])
        origin_dim   = len(origin_shape) - 1
        predict = sample['predict']
        input_shape = predict.shape
        scale = [(origin_shape[1:][i] + 0.0)/input_shape[2:][i] for \
                i in range(origin_dim)]
        scale = [1.0, 1.0] + scale

        output_predict = ndimage.interpolation.zoom(predict, scale, order = 1)
        sample['predict'] = output_predict
        return sample

class RandomRescale(AbstractTransform):
    """Rescale the image in a sample to a given size."""
    def __init__(self, params):
        """
        ratio0 (tuple/list or int): Desired minimal rescale ratio. 
            If tuple/list, the length should be 3 or 2.
        ratio1 (tuple/list or int): Desired maximal rescale ratio. 
            If tuple/list, the length should be 3 or 2.
        """
        super(RandomRescale, self).__init__(params)
        self.ratio0 = params["RandomRescale_lower_bound".lower()]
        self.ratio1 = params["RandomRescale_upper_bound".lower()]
        self.inverse     = params.get("RandomRescale_inverse".lower(), True)
        assert isinstance(self.ratio0, (float, list, tuple))
        assert isinstance(self.ratio1, (float, list, tuple))

    def __call__(self, sample):
        image = sample['image']
        input_shape = image.shape
        input_dim   = len(input_shape) - 1

        if isinstance(self.ratio0, (list, tuple)):
            for i in range(len(self.ratio0)):
                assert(self.ratio0[i] <= self.ratio1[i])
            scale = [self.ratio0[i] + random.random()*(self.ratio1[i] - self.ratio0[i]) \
                for i in range(len(self.ratio0))]
        else:
            scale = [self.ratio0 + random.random()*(self.ratio1 - self.ratio0) \
                for i in range(input_dim)]
        scale = [1.0] + scale
        image_t = ndimage.interpolation.zoom(image, scale, order = 1)

        sample['image'] = image_t
        sample['RandomRescale_origin_shape'] = json.dumps(input_shape)
        if('label' in sample and self.task == 'segmentation'):
            label = sample['label']
            label = ndimage.interpolation.zoom(label, scale, order = 0)
            sample['label'] = label
        if('pixel_weight' in sample and self.task == 'segmentation'):
            weight = sample['pixel_weight']
            weight = ndimage.interpolation.zoom(weight, scale, order = 1)
            sample['pixel_weight'] = weight
        
        return sample

    def inverse_transform_for_prediction(self, sample):
        ''' rescale sample['predict'] (5D or 4D) to the original spatial shape.
        origin_shape is a 4D or 3D vector as saved in __call__().'''
        if(isinstance(sample['RandomRescale_origin_shape'], list) or \
            isinstance(sample['RandomRescale_origin_shape'], tuple)):
            origin_shape = json.loads(sample['RandomRescale_origin_shape'][0])
        else:
            origin_shape = json.loads(sample['RandomRescale_origin_shape'])
        origin_dim   = len(origin_shape) - 1
        predict = sample['predict']
        input_shape = predict.shape
        scale = [(origin_shape[1:][i] + 0.0)/input_shape[2:][i] for \
                i in range(origin_dim)]
        scale = [1.0, 1.0] + scale

        output_predict = ndimage.interpolation.zoom(predict, scale, order = 1)
        sample['predict'] = output_predict
        return sample


# -*- coding: utf-8 -*-
from __future__ import print_function, division

import os
import torch
import pandas as pd
import numpy as np
from torch.utils.data import Dataset, DataLoader
from torchvision import transforms, utils
from pymic.io.image_read_write import load_image_as_nd_array

class NiftyDataset(Dataset):
    """Dataset for loading images. It generates 4D tensors with
    dimention order [C, D, H, W] for 3D images, and 3D tensors 
    with dimention order [C, H, W] for 2D images"""

    def __init__(self, root_dir, csv_file, modal_num = 1, 
            with_label = False, transform=None):
        """
        Args:
            root_dir (string): Directory with all the images.
            csv_file (string): Path to the csv file with image names.
            modal_num (int): Number of modalities. 
            with_label (bool): Load the data with segmentation ground truth.
            with_weight(bool): Load pixel-wise weight map.
            transform (callable, optional): Optional transform to be applied
                on a sample.
        """
        self.root_dir   = root_dir
        self.csv_items  = pd.read_csv(csv_file)
        self.modal_num  = modal_num
        self.with_label = with_label
        self.transform  = transform
       
        csv_keys = list(self.csv_items.keys())
        self.image_weight_idx = None
        self.pixel_weight_idx = None
        if('image_weight' in csv_keys):
            self.image_weight_idx = csv_keys.index('image_weight')
        if('pixel_weight' in csv_keys):
            self.pixel_weight_idx = csv_keys.index('pixel_weight')

    def __len__(self):
        return len(self.csv_items)

    def __getlabel__(self, idx):
        csv_keys = list(self.csv_items.keys())
        label_idx = csv_keys.index('label')
        label_name = "{0:}/{1:}".format(self.root_dir, 
            self.csv_items.iloc[idx, label_idx])
        label = load_image_as_nd_array(label_name)['data_array']
        label = np.asarray(label, np.int32)
        return label

    def __get_pixel_weight__(self, idx):
        weight_name = "{0:}/{1:}".format(self.root_dir, 
            self.csv_items.iloc[idx, self.pixel_weight_idx])
        weight = load_image_as_nd_array(weight_name)['data_array']
        weight = np.asarray(weight, np.float32)
        return weight        

    def __getitem__(self, idx):
        names_list, image_list = [], []
        for i in range (self.modal_num):
            image_name = self.csv_items.iloc[idx, i]
            image_full_name = "{0:}/{1:}".format(self.root_dir, image_name)
            image_dict = load_image_as_nd_array(image_full_name)
            image_data = image_dict['data_array']
            names_list.append(image_name)
            image_list.append(image_data)
        image = np.concatenate(image_list, axis = 0)
        image = np.asarray(image, np.float32)    
        sample = {'image': image, 'names' : names_list[0], 
                 'origin':image_dict['origin'],
                 'spacing': image_dict['spacing'],
                 'direction':image_dict['direction']}
        if (self.with_label):   
            sample['label'] = self.__getlabel__(idx) 
            assert(image.shape[1:] == sample['label'].shape[1:])
        if (self.image_weight_idx is not None):
            sample['image_weight'] = self.csv_items.iloc[idx, self.image_weight_idx]
        if (self.pixel_weight_idx is not None):
            sample['pixel_weight'] = self.__get_pixel_weight__(idx) 
            assert(image.shape[1:] == sample['pixel_weight'].shape[1:])
        if self.transform:
            sample = self.transform(sample)

        return sample


class ClassificationDataset(NiftyDataset):
    def __init__(self, root_dir, csv_file, modal_num = 1, class_num = 2, 
            with_label = False, transform=None):
        super(ClassificationDataset, self).__init__(root_dir, 
            csv_file, modal_num, with_label, transform)
        self.class_num = class_num

    def __getlabel__(self, idx):
        csv_keys = list(self.csv_items.keys())
        label_idx = csv_keys.index('label')
        label = self.csv_items.iloc[idx, label_idx]
        return label
    
    def __getweight__(self, idx):
        weight = self.csv_items.iloc[idx, self.image_weight_idx]
        weight = weight + 0.0
        return weight   
    
    def __getitem__(self, idx):
        names_list, image_list = [], []
        for i in range (self.modal_num):
            image_name = self.csv_items.iloc[idx, i]
            image_full_name = "{0:}/{1:}".format(self.root_dir, image_name)
            image_dict = load_image_as_nd_array(image_full_name)
            image_data = image_dict['data_array']
            names_list.append(image_name)
            image_list.append(image_data)
        image = np.concatenate(image_list, axis = 0)
        image = np.asarray(image, np.float32)    
        sample = {'image': image, 'names' : names_list[0], 
                 'origin':image_dict['origin'],
                 'spacing': image_dict['spacing'],
                 'direction':image_dict['direction']}
        if (self.with_label):   
            sample['label'] = self.__getlabel__(idx) 
        if (self.image_weight_idx is not None):
            sample['image_weight'] = self.__getweight__(idx) 
        if self.transform:
            sample = self.transform(sample)
        return sample



# -*- coding: utf-8 -*-
from __future__ import print_function, division

import torch
import json
import math
import random
import numpy as np
from scipy import ndimage
from pymic.transform.abstract_transform import AbstractTransform
from pymic.util.image_process import *


class GrayscaleToRGB(AbstractTransform):
    """
    apply random gamma correction to each channel
    """
    def __init__(self, params):
        """
        (gamma_min, gamma_max) specify the range of gamma
        """
        super(GrayscaleToRGB, self).__init__(params)
        self.inverse = params.get('GrayscaleToRGB_inverse'.lower(), False)
    
    def __call__(self, sample):
        image= sample['image']
        assert(image.shape[0] == 1 or image.shape[0] == 3)
        if(image.shape[0] == 1):
            sample['image'] = np.concatenate([image, image, image])
        return sample




# -*- coding: utf-8 -*-
from __future__ import print_function, division

import torch
import torch.nn as nn
from pymic.loss.cls.util import get_soft_label

class MSELoss(nn.Module):
    def __init__(self, params):
        super(MSELoss, self).__init__()
        self.mse_loss = nn.MSELoss()
    
    def forward(self, loss_input_dict):
        predict = loss_input_dict['prediction']
        labels  = loss_input_dict['ground_truth'][:, None] # reshape to N, 1
        softmax = nn.Softmax(dim = 1)
        predict = softmax(predict)
        num_class  = list(predict.size())[1]
        data_type = 'float' if(predict.dtype is torch.float32) else 'double'
        soft_y = get_soft_label(labels, num_class, data_type)
        loss = self.mse_loss(predict, soft_y)
        return loss



# -*- coding: utf-8 -*-
from __future__ import print_function, division
from pymic.net.net2d.unet2d import UNet2D
from pymic.net.net2d.unet2d_dual_branch import UNet2D_DualBranch
from pymic.net.net2d.unet2d_urpc import UNet2D_URPC
from pymic.net.net2d.unet2d_cct import UNet2D_CCT
from pymic.net.net2d.cople_net import COPLENet
from pymic.net.net2d.unet2d_attention import AttentionUNet2D
from pymic.net.net2d.unet2d_nest import NestedUNet2D
from pymic.net.net2d.unet2d_scse import UNet2D_ScSE
from pymic.net.net3d.unet2d5 import UNet2D5
from pymic.net.net3d.unet3d import UNet3D
from pymic.net.net3d.unet3d_scse import UNet3D_ScSE

SegNetDict = {
	'UNet2D': UNet2D,
	'UNet2D_DualBranch': UNet2D_DualBranch,
	'UNet2D_URPC': UNet2D_URPC,
	'UNet2D_CCT': UNet2D_CCT,
	'COPLENet': COPLENet,
	'AttentionUNet2D': AttentionUNet2D,
	'NestedUNet2D': NestedUNet2D,
	'UNet2D_ScSE': UNet2D_ScSE,
	'UNet2D5': UNet2D5,
	'UNet3D': UNet3D,
	'UNet3D_ScSE': UNet3D_ScSE
	}



import torch
import torch.nn as nn

class MSELoss(nn.Module):
    def __init__(self, params):
        super(MSELoss, self).__init__()
        if(params is None):
            self.softmax = True
        else:
            self.softmax = params.get('loss_softmax', True)
            
    def forward(self, loss_input_dict):
        predict = loss_input_dict['prediction']
        soft_y  = loss_input_dict['ground_truth']
        
        if(isinstance(predict, (list, tuple))):
            predict = predict[0]
        if(self.softmax):
            predict = nn.Softmax(dim = 1)(predict)
        mse  = torch.square(predict - soft_y)
        mse  = torch.mean(mse) 
        return mse 
    
    
class MAELoss(nn.MSELoss):
    def __init__(self, params):
        super(MAELoss, self).__init__(params)
        if(params is None):
            self.softmax = True
        else:
            self.softmax = params.get('loss_softmax', True)
    
    def get_prediction_error(self, loss_input_dict):
        predict = loss_input_dict['prediction']
        soft_y  = loss_input_dict['ground_truth']
        
        if(isinstance(predict, (list, tuple))):
            predict = predict[0]
        if(self.softmax):
            predict = nn.Softmax(dim = 1)(predict)
        mae = torch.abs(predict - soft_y)
        mae = torch.mean(mae)
        return mae 



# -*- coding: utf-8 -*-
from __future__ import print_function, division

import numpy as np
import torch
import torch.nn as nn
import SimpleITK as sitk
class SpaceToChannel3D(nn.Module):
    def __init__(self):
        super(SpaceToChannel3D, self).__init__()

    def forward(self, x):
        # only 3D images (5D tensor is support)
        input_shape = list(x.shape)
        assert(len(input_shape) == 5)
        [B,C, D, H, W] = input_shape
        assert((D % 2 == 0) and (H % 2 == 0) and (W % 2 == 0))
        halfD = int(D/2)
        halfH = int(H/2)
        halfW = int(W/2)
        # split along D axis
        x1 = x.contiguous().view([B, C, halfD, 2, H, W])
        # permute to [B, C, 2, halfD, H, W]
        x2 = x1.permute(0, 1, 3, 2, 4, 5)   
        # view as [B, 2*C, halfD, H, W] and [B, C*2, halfD, halfH, 2, W]
        x3 = x2.contiguous().view([B, C*2, halfD, halfH, 2, W])
        # permute to [B, C*2, 2, halfD, halfH, W]
        x4 = x3.permute(0, 1, 4, 2, 3, 5)  
        # view as [B, C*4, halfD, halfH, W] and [B, C*4, halfD, halfH, halfW, 2]
        x5 = x4.contiguous().view([B, C*4, halfD, halfH, halfW, 2])
        # permute to [B, C*4, 2, halfD, halfH, halfW]
        x6 = x5.permute(0, 1, 5, 2, 3, 4)
        x7 = x6.contiguous().view([B, C*8, halfD, halfH, halfW])
        return x7

class ChannelToSpace3D(nn.Module):
    def __init__(self):
        super(ChannelToSpace3D, self).__init__()

    def forward(self, x):
        # only 3D images (5D tensor is support)
        input_shape = list(x.shape)
        assert(len(input_shape) == 5)
        [B,C, D, H, W] = input_shape
        assert(C % 8 == 0)
        Cd8 = int(C/8)
        Cd4 = 2 * Cd8
        Cd2 = 2 * Cd4
        x6 = x.contiguous().view([B, Cd2, 2, D, H, W])
        # permute to [B, Cd4, D, H, W, 2]
        x5 = x6.permute(0, 1, 3, 4, 5, 2)
        x4 = x5.contiguous().view([B, Cd4, 2, D, H, 2*W])
        # permute to [B, Cd2, D, H, 2, 2*W]
        x3 = x4.permute(0, 1, 3, 4, 2, 5)
        x2 = x3.contiguous().view([B, Cd8, 2, D, 2* H, 2*W])
        x1 = x2.permute(0, 1, 3, 2, 4, 5)
        x0 = x1.contiguous().view([B, Cd8, 2*D, 2* H, 2*W])
        return x0


if __name__ == "__main__":
    s2c = SpaceToChannel3D()
    s2c = s2c.double()

    c2s = ChannelToSpace3D()
    c2s = c2s.double()

    img_name = "/home/disk2t/data/brats/BraTS2018_Training/HGG/Brats18_2013_2_1/Brats18_2013_2_1_flair.nii.gz"
    img_obj = sitk.ReadImage(img_name)
    img_data = sitk.GetArrayFromImage(img_obj)
    img_data = img_data[:-1]
    print(img_data.shape)
    x = img_data.reshape([1, 1] + list(img_data.shape))
    # x  = np.random.rand(4, 4, 96, 96, 96)
    xt = torch.from_numpy(x)
    xt = torch.tensor(xt)
    
    y = s2c(xt)
    z = c2s(y)
    y = y.detach().numpy()[0]
    print(y.shape)
    for i in range(8):
        sub_img = sitk.GetImageFromArray(y[i])
        # sub_img.CopyInformation(img_obj)
        save_name = "../../temp/{0:}.nii.gz".format(i)
        sitk.WriteImage(sub_img, save_name)
    z = z.detach().numpy()[0]
    print(z.shape)
    rec_img = sitk.GetImageFromArray(z[0])
    save_name = "../../temp/rec.nii.gz"
    sitk.WriteImage(rec_img, save_name)


# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function

import os
import numpy as np 
import SimpleITK as sitk 
from pymic.util.image_process import get_largest_k_components

class PostProcess(object):
    def __init__(self, params):
        self.params = params

    def __call__(self, seg):
        return seg

class PostKeepLargestComponent(PostProcess):
    def __init__(self, params):
        super(PostKeepLargestComponent, self).__init__(params)
        self.mode = params.get("KeepLargestComponent_mode".lower(), 1)
        """
        mode = 1: keep the largest component of the union of foreground classes. 
        mode = 2: keep the largest component for each foreground class.
        """

    def __call__(self, seg):
        if(self.mode == 1):
            mask = np.asarray(seg > 0, np.uint8)
            mask = get_largest_k_components(mask)
            seg = seg * mask
        elif(self.mode == 2):
            class_num = seg.max()
            output    = np.zeros_like(seg)
            for c in range(1, class_num + 1):
                seg_c  = np.asarray(seg == c, np.uint8)
                seg_c  = get_largest_k_components(seg_c) 
                output = output + seg_c * c
        return seg

PostProcessDict = {
    'KeepLargestComponent': PostKeepLargestComponent}


# -*- coding: utf-8 -*-
from __future__ import print_function, division

import torch
import torch.nn as nn


def get_acti_func(acti_func, params):
    acti_func = acti_func.lower()
    if(acti_func == 'relu'):
        inplace = params.get('relu_inplace', False)
        return nn.ReLU(inplace)

    elif(acti_func == 'leakyrelu'):
        slope   = params.get('leakyrelu_negative_slope', 1e-2)
        inplace = params.get('leakyrelu_inplace', False)
        return nn.LeakyReLU(slope, inplace)

    elif(acti_func == 'prelu'):
        num_params = params.get('prelu_num_parameters', 1)
        init_value = params.get('prelu_init', 0.25)
        return nn.PReLU(num_params, init_value)

    elif(acti_func == 'rrelu'):
        lower   = params.get('rrelu_lower', 1.0 /8)
        upper   = params.get('rrelu_upper', 1.0 /3)
        inplace = params.get('rrelu_inplace', False)
        return nn.RReLU(lower, upper, inplace)

    elif(acti_func == 'elu'):
        alpha   = params.get('elu_alpha', 1.0)
        inplace = params.get('elu_inplace', False)
        return nn.ELU(alpha, inplace)

    elif(acti_func == 'celu'):
        alpha   = params.get('celu_alpha', 1.0)
        inplace = params.get('celu_inplace', False)
        return nn.CELU(alpha, inplace)

    elif(acti_func == 'selu'):
        inplace = params.get('selu_inplace', False)
        return nn.SELU(inplace)

    elif(acti_func == 'glu'):
        dim = params.get('glu_dim', -1)
        return nn.GLU(dim)

    elif(acti_func == 'sigmoid'):
        return nn.Sigmoid()

    elif(acti_func == 'logsigmoid'):
        return nn.LogSigmoid()

    elif(acti_func == 'tanh'):
        return nn.Tanh()

    elif(acti_func == 'hardtanh'):
        min_val = params.get('hardtanh_min_val', -1.0)
        max_val = params.get('hardtanh_max_val',  1.0)
        inplace = params.get('hardtanh_inplace', False)
        return nn.Hardtanh(min_val, max_val, inplace)
    
    elif(acti_func == 'softplus'):
        beta      = params.get('softplus_beta', 1.0)
        threshold = params.get('softplus_threshold', 20)
        return nn.Softplus(beta, threshold)
    
    elif(acti_func == 'softshrink'):
        lambd = params.get('softshrink_lambda', 0.5)
        return nn.Softshrink(lambd)
    
    elif(acti_func == 'softsign'):
        return nn.Softsign()
    
    else:
        raise ValueError("Not implemented: {0:}".format(acti_func))


# -*- coding: utf-8 -*-
from __future__ import print_function, division
import logging
import os
import sys
from pymic.util.parse_config import *
from pymic.net_run.agent_cls import ClassificationAgent
from pymic.net_run.agent_seg import SegmentationAgent

def main():
    if(len(sys.argv) < 3):
        print('Number of arguments should be 3. e.g.')
        print('   pymic_run train config.cfg')
        exit()
    stage    = str(sys.argv[1])
    cfg_file = str(sys.argv[2])
    config   = parse_config(cfg_file)
    config   = synchronize_config(config)
    log_dir  = config['training']['ckpt_save_dir']
    if(not os.path.exists(log_dir)):
        os.makedirs(log_dir, exist_ok=True)
    logging.basicConfig(filename=log_dir+"/log_{0:}.txt".format(stage), level=logging.INFO,
                        format='%(message)s')
    logging.getLogger().addHandler(logging.StreamHandler(sys.stdout))
    logging_config(config)
    task     = config['dataset']['task_type']
    assert task in ['cls', 'cls_nexcl', 'seg']
    if(task == 'cls' or task == 'cls_nexcl'):
        agent = ClassificationAgent(config, stage)
    else:
        agent = SegmentationAgent(config, stage)
    agent.run()

if __name__ == "__main__":
    main()
    




# pretrained models from pytorch: https://pytorch.org/vision/0.8/models.html
from __future__ import print_function, division

import torch
import torch.nn as nn
import torchvision.models as models

# buildin_net_dict={
#     'resnet18': models.resnet18,
#     'alexnet':  models.alexnet,
#     'vgg16':    models.vgg16,
#     'squeezenet': models.squeezenet1_0,
#     'densenet': models.densenet161,
#     'inception':models.inception_v3,
#     'googlenet':models.googlenet,
#     'shufflenet': models.shufflenet_v2_x1_0,
#     'mobilenet':models.mobilenet,
#     'resnext50':models.resnext50_32x4d,
#     'wide_resnet': models.wide_resnet50_2,
#     'mnasnet': models.mnasnet1_0
# }

class ResNet18(nn.Module):
    def __init__(self, params):
        super(ResNet18, self).__init__()
        self.params    = params
        cls_num  = params['class_num']
        in_chns  = params.get('input_chns', 3)
        self.pretrain = params.get('pretrain', True)
        self.update_layers = params.get('update_layers', 0)
        self.net = models.resnet18(pretrained = self.pretrain)
        
        # replace the last layer 
        num_ftrs = self.net.fc.in_features
        self.net.fc = nn.Linear(num_ftrs, cls_num)
    
    def forward(self, x):
        return self.net(x)
    
    def get_parameters_to_update(self):
        if(self.pretrain == False or self.update_layers == 0):
            return self.net.parameters()
        elif(self.update_layers == -1):
            return self.net.fc.parameters()
        else:
            raise(ValueError("update_layers can only be 0 (all layers) " +
                "or -1 (the last layer)"))

class VGG16(nn.Module):
    def __init__(self, params):
        super(VGG16, self).__init__()
        self.params    = params
        cls_num  = params['class_num']
        in_chns  = params.get('input_chns', 3)
        self.pretrain = params.get('pretrain', True)
        self.update_layers = params.get('update_layers', 0)
        self.net = models.vgg16(pretrained = self.pretrain)
        
        # replace the last layer 
        num_ftrs = self.net.classifier[-1].in_features
        self.net.classifier[-1] = nn.Linear(num_ftrs, cls_num)
    
    def forward(self, x):
        return self.net(x)
    
    def get_parameters_to_update(self):
        if(self.pretrain == False or self.update_layers == 0):
            return self.net.parameters()
        elif(self.update_layers == -1):
            return self.net.classifier[-1].parameters()
        else:
            raise(ValueError("update_layers can only be 0 (all layers) " +
                "or -1 (the last layer)"))

class MobileNetV2(nn.Module):
    def __init__(self, params):
        super(MobileNetV2, self).__init__()
        self.params = params
        cls_num  = params['class_num']
        in_chns  = params.get('input_chns', 3)
        self.pretrain = params.get('pretrain', True)
        self.update_layers = params.get('update_layers', 0)
        self.net = models.mobilenet_v2(pretrained = self.pretrain)
        
        # replace the last layer 
        num_ftrs = self.net.last_channel
        self.net.classifier[-1] = nn.Linear(num_ftrs, cls_num)
    
    def forward(self, x):
        return self.net(x)
    
    def get_parameters_to_update(self):
        if(self.pretrain == False or self.update_layers == 0):
            return self.net.parameters()
        elif(self.update_layers == -1):
            return self.net.classifier[-1].parameters()
        else:
            raise(ValueError("update_layers can only be 0 (all layers) " +
                "or -1 (the last layer)"))


# -*- coding: utf-8 -*-
from __future__ import print_function, division

import torch
import torch.nn as nn

class DeconvolutionLayer(nn.Module):
    """
    A compose layer with the following components:
    deconvolution -> (batch_norm) -> activation -> (dropout)
    batch norm and dropout are optional
    """
    def __init__(self, in_channels, out_channels, kernel_size, 
            dim = 3, stride = 1, padding = 0, output_padding = 0, 
            dilation =1, groups = 1, bias = True, 
            batch_norm = True, acti_func = None):
        super(DeconvolutionLayer, self).__init__()
        self.n_in_chns  = in_channels
        self.n_out_chns = out_channels
        self.batch_norm = batch_norm
        self.acti_func  = acti_func
        
        assert(dim == 2 or dim == 3)
        if(dim == 2):
            self.conv = nn.ConvTranspose2d(in_channels, out_channels,
                kernel_size, stride, padding, output_padding,
                groups, bias, dilation)
            if(self.batch_norm):
                self.bn = nn.BatchNorm2d(out_channels)
        else:
            self.conv = nn.ConvTranspose3d(in_channels, out_channels,
                kernel_size, stride, padding, output_padding,
                groups, bias, dilation)
            if(self.batch_norm):
                self.bn = nn.BatchNorm3d(out_channels)

    def forward(self, x):
        f = self.conv(x)
        if(self.batch_norm):
            f = self.bn(f)
        if(self.acti_func is not None):
            f = self.acti_func(f)
        return f

class  DepthSeperableDeconvolutionLayer(nn.Module):
    """
    A compose layer with the following components:
    convolution -> (batch_norm) -> activation -> (dropout)
    batch norm and dropout are optional
    """
    def __init__(self, in_channels, out_channels, kernel_size, 
        dim = 3, stride = 1, padding = 0, output_padding = 0,
        dilation =1, groups = 1, bias = True, 
        batch_norm = True, acti_func = None):
        super(DepthSeperableDeconvolutionLayer, self).__init__()
        self.n_in_chns  = in_channels
        self.n_out_chns = out_channels
        self.batch_norm = batch_norm
        self.acti_func  = acti_func
        self.groups     = groups
        assert(dim == 2 or dim == 3)
        if(dim == 2):
            self.conv1x1 = nn.Conv2d(in_channels, out_channels,
                kernel_size = 1, stride = 1, padding = 0, dilation = dilation, 
                groups = self.groups, bias = bias)
            self.conv = nn.ConvTranspose2d(out_channels, out_channels,
                kernel_size, stride, padding, output_padding,
                groups = out_channels, bias = bias, dilation = dilation)
            
            if(self.batch_norm):
                self.bn = nn.BatchNorm2d(out_channels)
        else:
            self.conv1x1 = nn.Conv3d(in_channels, out_channels,
                kernel_size = 1, stride = 1, padding = 0, dilation = dilation, 
                groups = self.groups, bias = bias)
            self.conv = nn.ConvTranspose3d(out_channels, out_channels,
                kernel_size, stride, padding, output_padding,
                groups = out_channels, bias = bias, dilation = dilation)
            if(self.batch_norm):
                self.bn = nn.BatchNorm3d(out_channels)

    def forward(self, x):
        f = self.conv1x1(x)
        f = self.conv(f)
        if(self.batch_norm):
            f = self.bn(f)
        if(self.acti_func is not None):
            f = self.acti_func(f)
        return f


# -*- coding: utf-8 -*-
from __future__ import print_function, division
import logging
import numpy as np
import random
import torch
import torch.nn.functional as F
from torch.optim import lr_scheduler
from pymic.loss.seg.util import get_soft_label
from pymic.loss.seg.util import reshape_prediction_and_ground_truth
from pymic.loss.seg.util import get_classwise_dice
from pymic.net.net_dict_seg import SegNetDict
from pymic.net_run_wsl.wsl_abstract import WSLSegAgent
from pymic.util.ramps import get_rampup_ratio
from pymic.util.general import keyword_match

class WSLUSTM(WSLSegAgent):
    """
    USTM for scribble-supervised segmentation according to the following paper:
        Xiaoming Liu, Quan Yuan, Yaozong Gao, Helei He, Shuo Wang, Xiao Tang,
        Jinshan Tang, Dinggang Shen:
        Weakly Supervised Segmentation of COVID19 Infection with Scribble Annotation on CT Images.
        Patter Recognition, 2022.
        https://doi.org/10.1016/j.patcog.2021.108341 
    """
    def __init__(self, config, stage = 'train'):
        super(WSLUSTM, self).__init__(config, stage)
        self.net_ema = None 
    
    def create_network(self):
        super(WSLUSTM, self).create_network()
        if(self.net_ema is None):
            net_name = self.config['network']['net_type']
            if(net_name not in SegNetDict):
                raise ValueError("Undefined network {0:}".format(net_name))
            self.net_ema = SegNetDict[net_name](self.config['network'])
        if(self.tensor_type == 'float'):
            self.net_ema.float()
        else:
            self.net_ema.double()

    def training(self):
        class_num   = self.config['network']['class_num']
        iter_valid  = self.config['training']['iter_valid']
        wsl_cfg     = self.config['weakly_supervised_learning']
        iter_max     = self.config['training']['iter_max']
        rampup_start = wsl_cfg.get('rampup_start', 0)
        rampup_end   = wsl_cfg.get('rampup_end', iter_max)
        train_loss  = 0
        train_loss_sup = 0
        train_loss_reg = 0
        train_dice_list = []
        self.net.train()
        self.net_ema.to(self.device)
        for it in range(iter_valid):
            try:
                data = next(self.trainIter)
            except StopIteration:
                self.trainIter = iter(self.train_loader)
                data = next(self.trainIter)
            
            # get the inputs
            inputs = self.convert_tensor_type(data['image'])
            y      = self.convert_tensor_type(data['label_prob'])  
                         
            inputs, y = inputs.to(self.device), y.to(self.device)
            
            # zero the parameter gradients
            self.optimizer.zero_grad()
                
            # forward + backward + optimize
            noise   = torch.clamp(torch.randn_like(inputs) * 0.1, -0.2, 0.2)
            outputs = self.net(inputs + noise)
            out_prob= F.softmax(outputs, dim=1)
            loss_sup = self.get_loss_value(data, outputs, y)

            rot_times = random.randrange(0, 4)
            inputs_rot= torch.rot90(inputs, rot_times, [-2, -1])
            noise = torch.clamp(torch.randn_like(inputs_rot) * 0.1, -0.2, 0.2)
            with torch.no_grad():
                ema_inputs  = inputs_rot + noise
                ema_outputs = self.net_ema(ema_inputs)
                ema_out_prob= F.softmax(ema_outputs, dim=1)
            out_prob_rot = torch.rot90(out_prob, rot_times, [-2, -1])
            square_error = torch.square(out_prob_rot - ema_out_prob)

            # the forward pass number for uncertainty estimation
            T = wsl_cfg.get("ustm_mcdroput_n", 8)
            preds = torch.zeros([T] + list(y.shape)).to(self.device)
            for i in range(T//2):
                ema_inputs_r = torch.cat([inputs_rot, inputs_rot], dim = 0)
                ema_inputs_r = ema_inputs_r + \
                    torch.clamp(torch.randn_like(ema_inputs_r) * 0.1, -0.2, 0.2)
                ema_inputs_r = ema_inputs_r.to(self.device)
                with torch.no_grad():
                    ema_outputs_r = self.net_ema(ema_inputs_r)
                # reshape from [2B, C, D, H, W] to [2, B, C, D, H, W]
                preds[2*i:2*(i+1)] = ema_outputs_r.reshape([2]+list(y.shape))
            preds = torch.softmax(preds, dim = 2)
            preds = torch.mean(preds, dim = 0)
            uncertainty = -1.0 * torch.sum(preds*torch.log(preds + 1e-6),
                 dim=1, keepdim=True)
            
            rampup_ratio = get_rampup_ratio(self.glob_it, rampup_start, rampup_end, "sigmoid")
            class_num = list(y.shape)[1]
            threshold = (0.75+0.25*rampup_ratio)*np.log(class_num)
            mask      = (uncertainty < threshold).float()
            loss_reg  = torch.sum(mask*square_error)/(2*torch.sum(mask)+1e-16)

            regular_w = wsl_cfg.get('regularize_w', 0.1) * rampup_ratio
            loss = loss_sup + regular_w*loss_reg

            loss.backward()
            self.optimizer.step()
            if(self.scheduler is not None and \
                not isinstance(self.scheduler, lr_scheduler.ReduceLROnPlateau)):
                self.scheduler.step()

            # update EMA
            alpha = wsl_cfg.get('ema_decay', 0.99)
            alpha = min(1 - 1 / (iter_max + 1), alpha)
            for ema_param, param in zip(self.net_ema.parameters(), self.net.parameters()):
                ema_param.data.mul_(alpha).add_(1 - alpha, param.data)

            train_loss = train_loss + loss.item()
            train_loss_sup = train_loss_sup + loss_sup.item()
            train_loss_reg = train_loss_reg + loss_reg.item() 
            # get dice evaluation for each class in annotated images
            if(isinstance(outputs, tuple) or isinstance(outputs, list)):
                outputs = outputs[0] 
            p_argmax = torch.argmax(outputs, dim = 1, keepdim = True)
            p_soft   = get_soft_label(p_argmax, class_num, self.tensor_type)
            p_soft, y = reshape_prediction_and_ground_truth(p_soft, y) 
            dice_list   = get_classwise_dice(p_soft, y)
            train_dice_list.append(dice_list.cpu().numpy())
        train_avg_loss = train_loss / iter_valid
        train_avg_loss_sup = train_loss_sup / iter_valid
        train_avg_loss_reg = train_loss_reg / iter_valid
        train_cls_dice = np.asarray(train_dice_list).mean(axis = 0)
        train_avg_dice = train_cls_dice.mean()

        train_scalers = {'loss': train_avg_loss, 'loss_sup':train_avg_loss_sup,
            'loss_reg':train_avg_loss_reg, 'regular_w':regular_w,
            'avg_dice':train_avg_dice,     'class_dice': train_cls_dice}
        return train_scalers


# -*- coding: utf-8 -*-
from __future__ import print_function, division

import torch
import json
import math
import random
import numpy as np
from scipy import ndimage
from pymic.transform.abstract_transform import AbstractTransform
from pymic.util.image_process import *


class ChannelWiseThreshold(AbstractTransform):
    """Threshold the image (shape [C, D, H, W] or [C, H, W]) for each channel
    """
    def __init__(self, params):
        """
        channels (tuple/list/None): the list of specified channels for thresholding. Default value 
            is all the channels.
        threshold_lower (tuple/list/None): The lower threshold values for specified channels.
        threshold_upper (tuple/list/None): The uppoer threshold values for specified channels.
        replace_lower (tuple/list/None): new values for pixels with intensity smaller than 
            threshold_lower. Default value is 
        replace_upper (tuple/list/None): new values for pixels with intensity larger than threshold_upper.
        """
        super(ChannelWiseThreshold, self).__init__(params)
        self.channlels = params['ChannelWiseThreshold_channels'.lower()]
        self.threshold_lower = params['ChannelWiseThreshold_threshold_lower'.lower()]
        self.threshold_upper = params['ChannelWiseThreshold_threshold_upper'.lower()]
        self.replace_lower   = params['ChannelWiseThreshold_replace_lower'.lower()]
        self.replace_upper   = params['ChannelWiseThreshold_replace_upper'.lower()]
        self.inverse = params.get('ChannelWiseThreshold_inverse'.lower(), False)

    def __call__(self, sample):
        image= sample['image']
        channels = range(image.shape[0]) if self.channlels is None else self.channlels
        for i in range(len(channels)):
            chn = channels[i]
            if((self.threshold_lower is not None) and (self.threshold_lower[i] is not None)):
                t_lower = self.threshold_lower[i]
                r_lower = self.threshold_lower[i]
                if((self.replace_lower is not None) and (self.replace_lower[i] is not None)):
                    r_lower = self.replace_lower[i]
                image[chn][image[chn] < t_lower] = r_lower
            
            if((self.threshold_upper is not None) and (self.threshold_upper[i] is not None)):
                t_upper = self.threshold_upper[i]
                r_upper = self.threshold_upper[i]
                if((self.replace_upper is not None) and (self.replace_upper[i] is not None)):
                    r_upper= self.replace_upper[i]
                image[chn][image[chn] > t_upper] = r_upper
        sample['image'] = image
        return sample

class ChannelWiseThresholdWithNormalize(AbstractTransform):
    """
    Note that this can be replaced by ChannelWiseThreshold + NormalizeWithMinMax
    
    Threshold the image (shape [C, D, H, W] or [C, H, W]) for each channel
       and then normalize the image based on remaining pixels
    """
    def __init__(self, params):
        """
        :param threshold_lower: (tuple/list/None) The lower threshold value along each channel.
        :param threshold_upper: (typle/list/None) The upper threshold value along each channel.
        :param mean_std_mode: (bool) If true, nomalize the image based on mean and std values,
            and pixels values outside the threshold value are replaced random number.
            If false, use the min and max values for normalization.
        """
        super(ChannelWiseThresholdWithNormalize, self).__init__(params)
        self.threshold_lower = params['ChannelWiseThresholdWithNormalize_threshold_lower'.lower()]
        self.threshold_upper = params['ChannelWiseThresholdWithNormalize_threshold_upper'.lower()]
        self.mean_std_mode   = params['ChannelWiseThresholdWithNormalize_mean_std_mode'.lower()]
        self.inverse = params.get('ChannelWiseThresholdWithNormalize_inverse'.lower(), False)

    def __call__(self, sample):
        image= sample['image']
        for chn in range(image.shape[0]):
            v0 = self.threshold_lower[chn]
            v1 = self.threshold_upper[chn]
            if(self.mean_std_mode == True):
                mask = np.ones_like(image[chn])
                if(v0 is not None):
                    mask = mask * np.asarray(image[chn] > v0)
                if(v1 is not None):
                    mask = mask * np.asarray(image[chn] < v1)
                pixels   = image[chn][mask > 0]
                chn_mean = pixels.mean()
                chn_std  = pixels.std()
                chn_norm = (image[chn] - chn_mean)/chn_std
                chn_random = np.random.normal(0, 1, size = chn_norm.shape)
                chn_norm[mask == 0] = chn_random[mask == 0]
                image[chn] = chn_norm
            else:
                img_chn = image[chn]
                if(v0 is not None):
                    img_chn[img_chn < v0] = v0
                    min_value = v0 
                else:
                    min_value = img_chn.min()
                if(v1 is not None):
                    img_chn[img_chn > v1] = v1 
                    max_value = img_chn.max() 
                else:
                    max_value = img_chn.max() 
                img_chn = (img_chn - min_value) / (max_value - min_value)
                image[chn] = img_chn
        sample['image'] = image
        return sample


# -*- coding: utf-8 -*-
"""
Combining 3D U-Net with SCSE module according to the following paper:
    Abhijit Guha Roy, Nassir Navab, Christian Wachinger:
    Recalibrating Fully Convolutional Networks With Spatial and Channel "Squeeze and Excitation" Blocks. \
    IEEE Trans. Med. Imaging 38(2): 540-549 (2019)
"""
from __future__ import print_function, division

import torch
import torch.nn as nn
import numpy as np
from pymic.net.net3d.scse3d import *

class ConvScSEBlock3D(nn.Module):
    """two convolution layers with batch norm and leaky relu"""
    def __init__(self,in_channels, out_channels, dropout_p):
        super(ConvScSEBlock3D, self).__init__()
        self.conv_conv = nn.Sequential(
            nn.Conv3d(in_channels, out_channels, kernel_size=3, padding=1),
            nn.BatchNorm3d(out_channels),
            nn.LeakyReLU(),
            nn.Dropout(dropout_p),
            nn.Conv3d(out_channels, out_channels, kernel_size=3, padding=1),
            nn.BatchNorm3d(out_channels),
            nn.LeakyReLU(),
            ChannelSpatialSELayer3D(out_channels)
        )
       
    def forward(self, x):
        return self.conv_conv(x)

class DownBlock(nn.Module):
    """Downsampling followed by ConvBlock"""
    def __init__(self, in_channels, out_channels, dropout_p):
        super(DownBlock, self).__init__()
        self.maxpool_conv = nn.Sequential(
            nn.MaxPool3d(2),
            ConvScSEBlock3D(in_channels, out_channels, dropout_p)
        )

    def forward(self, x):
        return self.maxpool_conv(x)

class UpBlock(nn.Module):
    """Upssampling followed by ConvBlock"""
    def __init__(self, in_channels1, in_channels2, out_channels, dropout_p, 
                 bilinear=True):
        super(UpBlock, self).__init__()
        self.bilinear = bilinear
        if bilinear:
            self.conv1x1 = nn.Conv3d(in_channels1, in_channels2, kernel_size = 1)
            self.up = nn.Upsample(scale_factor=2, mode='trilinear', align_corners=True)
        else:
            self.up = nn.ConvTranspose3d(in_channels1, in_channels2, kernel_size=2, stride=2)
        self.conv = ConvScSEBlock3D(in_channels2 * 2, out_channels, dropout_p)

    def forward(self, x1, x2):
        if self.bilinear:
            x1 = self.conv1x1(x1)
        x1 = self.up(x1)
        x = torch.cat([x2, x1], dim=1)
        return self.conv(x)

class UNet3D_ScSE(nn.Module):
    def __init__(self, params):
        super(UNet3D_ScSE, self).__init__()
        self.params    = params
        self.in_chns   = self.params['in_chns']
        self.ft_chns   = self.params['feature_chns']
        self.n_class   = self.params['class_num']
        self.bilinear  = self.params['trilinear']
        self.dropout   = self.params['dropout']
        assert(len(self.ft_chns) == 5)

        self.in_conv= ConvScSEBlock3D(self.in_chns, self.ft_chns[0], self.dropout[0])
        self.down1  = DownBlock(self.ft_chns[0], self.ft_chns[1], self.dropout[1])
        self.down2  = DownBlock(self.ft_chns[1], self.ft_chns[2], self.dropout[2])
        self.down3  = DownBlock(self.ft_chns[2], self.ft_chns[3], self.dropout[3])
        self.down4  = DownBlock(self.ft_chns[3], self.ft_chns[4], self.dropout[4])
        self.up1 = UpBlock(self.ft_chns[4], self.ft_chns[3], self.ft_chns[3], dropout_p = self.dropout[3]) 
        self.up2 = UpBlock(self.ft_chns[3], self.ft_chns[2], self.ft_chns[2], dropout_p = self.dropout[2]) 
        self.up3 = UpBlock(self.ft_chns[2], self.ft_chns[1], self.ft_chns[1], dropout_p = self.dropout[1]) 
        self.up4 = UpBlock(self.ft_chns[1], self.ft_chns[0], self.ft_chns[0], dropout_p = self.dropout[0]) 
    
        self.out_conv = nn.Conv3d(self.ft_chns[0], self.n_class,  
            kernel_size = 3, padding = 1)

    def forward(self, x):
        
        x0 = self.in_conv(x)
        x1 = self.down1(x0)
        x2 = self.down2(x1)
        x3 = self.down3(x2)
        x4 = self.down4(x3)
        
        x = self.up1(x4, x3)
        x = self.up2(x, x2)
        x = self.up3(x, x1)
        x = self.up4(x, x0)
        output = self.out_conv(x)

        return output

if __name__ == "__main__":
    params = {'in_chns':4,
              'feature_chns':[2, 8, 32, 48, 64],
              'dropout':  [0, 0, 0.3, 0.4, 0.5],
              'class_num': 2,
              'trilinear': True}
    Net = UNet3D_ScSE(params)
    Net = Net.double()

    x  = np.random.rand(4, 4, 96, 96, 96)
    xt = torch.from_numpy(x)
    xt = torch.tensor(xt)
    
    y = Net(xt)
    print(len(y.size()))
    y = y.detach().numpy()
    print(y.shape)


# -*- coding: utf-8 -*-
"""
Implementation of Co-teaching for learning from noisy samples for 
segmentation tasks according to the following paper:
    Bo Han et al., Co-teaching: Robust Training of Deep NeuralNetworks
    with Extremely Noisy Labels, NeurIPS, 2018
The author's original implementation was:
https://github.com/bhanML/Co-teaching 


"""
from __future__ import print_function, division
import logging
import os
import sys
import numpy as np
import torch
import torch.nn as nn
import torch.optim as optim
from torch.optim import lr_scheduler
from pymic.loss.seg.util import get_soft_label
from pymic.loss.seg.util import reshape_prediction_and_ground_truth
from pymic.loss.seg.util import get_classwise_dice
from pymic.loss.seg.util import reshape_tensor_to_2D
from pymic.net_run.agent_seg import SegmentationAgent
from pymic.net.net_dict_seg import SegNetDict
from pymic.util.parse_config import *
from pymic.util.ramps import get_rampup_ratio



class TriNet(nn.Module):
    def __init__(self, params):
        super(TriNet, self).__init__()
        net_name  = params['net_type']
        self.net1 = SegNetDict[net_name](params)
        self.net2 = SegNetDict[net_name](params)
        self.net3 = SegNetDict[net_name](params)    

    def forward(self, x):
        out1 = self.net1(x)
        out2 = self.net2(x)
        out3 = self.net3(x)

        if(self.training):
          return out1, out2, out3
        else:
          return (out1 + out2 + out3) / 3

class NLLTriNet(SegmentationAgent):
    """
    Co-teaching: Robust Training of Deep Neural Networks with Extremely 
    Noisy Labels
    https://arxiv.org/abs/1804.06872
    """
    def __init__(self, config, stage = 'train'):
        super(NLLTriNet, self).__init__(config, stage)
       
    def create_network(self):
        if(self.net is None):
            self.net = TriNet(self.config['network'])
        if(self.tensor_type == 'float'):
            self.net.float()
        else:
            self.net.double()

    def get_loss_and_confident_mask(self, pred, labels_prob, conf_ratio):
        prob = nn.Softmax(dim = 1)(pred)
        prob_2d = reshape_tensor_to_2D(prob) * 0.999 + 5e-4
        y_2d  = reshape_tensor_to_2D(labels_prob)

        loss = - y_2d* torch.log(prob_2d)
        loss = torch.sum(loss, dim = 1) # shape is [N]
        threshold   = torch.quantile(loss, conf_ratio)
        mask = loss < threshold
        return loss, mask

    def training(self):
        class_num   = self.config['network']['class_num']
        iter_valid  = self.config['training']['iter_valid']
        nll_cfg     = self.config['noisy_label_learning']
        iter_max     = self.config['training']['iter_max']
        select_ratio = nll_cfg['trinet_select_ratio']
        rampup_start = nll_cfg.get('rampup_start', 0)
        rampup_end   = nll_cfg.get('rampup_end', iter_max)

        train_loss_no_select1 = 0
        train_loss_no_select2 = 0
        train_loss1, train_loss2, train_loss3 = 0, 0, 0
        train_dice_list = []
        self.net.train()
        for it in range(iter_valid):
            try:
                data = next(self.trainIter)
            except StopIteration:
                self.trainIter = iter(self.train_loader)
                data = next(self.trainIter)
            
            # get the inputs
            inputs      = self.convert_tensor_type(data['image'])
            labels_prob = self.convert_tensor_type(data['label_prob'])
            inputs, labels_prob = inputs.to(self.device), labels_prob.to(self.device)

            # zero the parameter gradients
            self.optimizer.zero_grad()
                
            # forward + backward + optimize
            outputs1, outputs2, outputs3 = self.net(inputs)

            rampup_ratio = get_rampup_ratio(self.glob_it, rampup_start, rampup_end)
            forget_ratio = (1 - select_ratio) * rampup_ratio
            remb_ratio   = 1 - forget_ratio

            loss1, mask1 = self.get_loss_and_confident_mask(outputs1, labels_prob, remb_ratio)
            loss2, mask2 = self.get_loss_and_confident_mask(outputs2, labels_prob, remb_ratio)
            loss3, mask3 = self.get_loss_and_confident_mask(outputs3, labels_prob, remb_ratio)
            mask12, mask13, mask23 = mask1 * mask2, mask1 * mask3, mask2 * mask3 
            mask12, mask13, mask23 = mask12.detach(), mask13.detach(), mask23.detach()

            loss1_avg = torch.sum(loss1 * mask23) / mask23.sum()
            loss2_avg = torch.sum(loss2 * mask13) / mask13.sum()
            loss3_avg = torch.sum(loss3 * mask12) / mask12.sum()
            loss = (loss1_avg + loss2_avg + loss3_avg) / 3

            loss.backward()
            self.optimizer.step()
            if(self.scheduler is not None and \
                not isinstance(self.scheduler, lr_scheduler.ReduceLROnPlateau)):
                self.scheduler.step()

            train_loss_no_select1 = train_loss_no_select1 + loss1.mean().item()
            train_loss_no_select2 = train_loss_no_select2 + loss2.mean().item()
            train_loss1 = train_loss1 + loss1_avg.item()
            train_loss2 = train_loss2 + loss2_avg.item()

            outputs1_argmax = torch.argmax(outputs1, dim = 1, keepdim = True)
            soft_out1       = get_soft_label(outputs1_argmax, class_num, self.tensor_type)
            soft_out1, labels_prob = reshape_prediction_and_ground_truth(soft_out1, labels_prob)  
            dice_list   = get_classwise_dice(soft_out1, labels_prob).detach().cpu().numpy()
            train_dice_list.append(dice_list)
        train_avg_loss_no_select1 = train_loss_no_select1 / iter_valid
        train_avg_loss_no_select2 = train_loss_no_select2 / iter_valid
        train_avg_loss1 = train_loss1 / iter_valid
        train_avg_loss2 = train_loss2 / iter_valid
        train_cls_dice = np.asarray(train_dice_list).mean(axis = 0)
        train_avg_dice = train_cls_dice.mean()

        train_scalers = {'loss': (train_avg_loss1 + train_avg_loss2) / 2, 
            'loss1':train_avg_loss1, 'loss2': train_avg_loss2,
            'loss_no_select1':train_avg_loss_no_select1, 
            'loss_no_select2':train_avg_loss_no_select2,
            'select_ratio':remb_ratio, 'avg_dice':train_avg_dice, 'class_dice': train_cls_dice}
        return train_scalers
    
    def write_scalars(self, train_scalars, valid_scalars, lr_value, glob_it):
        loss_scalar ={'train':train_scalars['loss'], 
                      'valid':valid_scalars['loss']}
        loss_no_select_scalar  = {'net1':train_scalars['loss_no_select1'],
                                  'net2':train_scalars['loss_no_select2']}

        dice_scalar ={'train':train_scalars['avg_dice'], 'valid':valid_scalars['avg_dice']}
        self.summ_writer.add_scalars('loss', loss_scalar, glob_it)
        self.summ_writer.add_scalars('loss_no_select', loss_no_select_scalar, glob_it)
        self.summ_writer.add_scalars('select_ratio', {'select_ratio':train_scalars['select_ratio']}, glob_it)
        self.summ_writer.add_scalars('lr', {"lr": lr_value}, glob_it)
        self.summ_writer.add_scalars('dice', dice_scalar, glob_it)
        class_num = self.config['network']['class_num']
        for c in range(class_num):
            cls_dice_scalar = {'train':train_scalars['class_dice'][c], \
                'valid':valid_scalars['class_dice'][c]}
            self.summ_writer.add_scalars('class_{0:}_dice'.format(c), cls_dice_scalar, glob_it)

        logging.info('train loss {0:.4f}, avg dice {1:.4f} '.format(
            train_scalars['loss'], train_scalars['avg_dice']) + "[" + \
            ' '.join("{0:.4f}".format(x) for x in train_scalars['class_dice']) + "]")        
        logging.info('valid loss {0:.4f}, avg dice {1:.4f} '.format(
            valid_scalars['loss'], valid_scalars['avg_dice']) + "[" + \
            ' '.join("{0:.4f}".format(x) for x in valid_scalars['class_dice']) + "]") 



# -*- coding: utf-8 -*-
from __future__ import print_function, division

import torch
import json
import math
import random
import numpy as np
from scipy import ndimage
from pymic.transform.abstract_transform import AbstractTransform
from pymic.util.image_process import *


class RandomRotate(AbstractTransform):
    """
    random rotate the image (shape [C, D, H, W] or [C, H, W]) 
    """
    def __init__(self, params): 
        """
        angle_range_d (tuple/list/None) : rorate angle range along depth axis (degree),
               only used for 3D images
        angle_range_h (tuple/list/None) : rorate angle range along height axis (degree)
        angle_range_w (tuple/list/None) : rorate angle range along width axis (degree)
        """
        super(RandomRotate, self).__init__(params)
        self.angle_range_d  = params['RandomRotate_angle_range_d'.lower()]
        self.angle_range_h  = params['RandomRotate_angle_range_h'.lower()]
        self.angle_range_w  = params['RandomRotate_angle_range_w'.lower()]
        self.inverse = params.get('RandomRotate_inverse'.lower(), True)

    def __apply_transformation(self, image, transform_param_list, order = 1):
        """
        apply rotation transformation to an ND image
        Args:
            image (nd array): the input nd image
            transform_param_list (list): a list of roration angle and axes
            order (int): interpolation order
        """
        for angle, axes in transform_param_list:
            image = ndimage.rotate(image, angle, axes, reshape = False, order = order)
        return image

    def __call__(self, sample):
        image = sample['image']
        input_shape = image.shape
        input_dim = len(input_shape) - 1
        
        transform_param_list = []
        if(self.angle_range_d is not None):
            angle_d = np.random.uniform(self.angle_range_d[0], self.angle_range_d[1])
            transform_param_list.append([angle_d, (-1, -2)])
        if(input_dim == 3):
            if(self.angle_range_h is not None):
                angle_h = np.random.uniform(self.angle_range_h[0], self.angle_range_h[1])
                transform_param_list.append([angle_h, (-1, -3)])
            if(self.angle_range_w is not None):
                angle_w = np.random.uniform(self.angle_range_w[0], self.angle_range_w[1])
                transform_param_list.append([angle_w, (-2, -3)])
        assert(len(transform_param_list) > 0)

        sample['RandomRotate_Param'] = json.dumps(transform_param_list)
        image_t = self.__apply_transformation(image, transform_param_list, 1)
        sample['image'] = image_t
        if('label' in sample and self.task == 'segmentation'):
            sample['label'] = self.__apply_transformation(sample['label'] , 
                                transform_param_list, 0)
        if('pixel_weight' in sample and self.task == 'segmentation'):
            sample['pixel_weight'] = self.__apply_transformation(sample['pixel_weight'] , 
                                transform_param_list, 1)
        return sample

    def  inverse_transform_for_prediction(self, sample):
        ''' rorate sample['predict'] (5D or 4D) to the original direction.
        transform_param_list is a list as saved in __call__().'''
        # get the paramters for invers transformation
        if(isinstance(sample['RandomRotate_Param'], list) or \
            isinstance(sample['RandomRotate_Param'], tuple)):
            transform_param_list = json.loads(sample['RandomRotate_Param'][0]) 
        else:
            transform_param_list = json.loads(sample['RandomRotate_Param']) 
        transform_param_list.reverse()
        for i in range(len(transform_param_list)):
            transform_param_list[i][0] = - transform_param_list[i][0]
        sample['predict'] = self.__apply_transformation(sample['predict'] , 
                                transform_param_list, 1)
        return sample


# -*- coding: utf-8 -*-
"""
A 2.5D network combining 3D convolutions with 2D convolutions according to the following paper:
    Guotai Wang, Jonathan Shapey, Wenqi Li, Reuben Dorent, Alex Demitriadis, Sotirios Bisdas, 
    Ian Paddick, Robert Bradford, Shaoting Zhang, Sébastien Ourselin, Tom Vercauteren:
    Automatic Segmentation of Vestibular Schwannoma from T2-Weighted MRI by Deep Spatial Attention
    with Hardness-Weighted Loss. MICCAI (2) 2019: 264-272.
Note that the attention module in the orininal paper is not used here.
"""
from __future__ import print_function, division
import torch
import torch.nn as nn
import numpy as np

class ConvBlockND(nn.Module):
    """for 2D and 3D convolutional blocks"""
    def __init__(self, in_channels, out_channels, 
                dim = 2, dropout_p = 0.0):
        """
        dim: should be 2 or 3
        dropout_p: probability to be zeroed
        """
        super(ConvBlockND, self).__init__()
        assert(dim == 2 or dim == 3)
        self.dim = dim 
        if(self.dim == 2):
            self.conv_conv = nn.Sequential(
                nn.Conv2d(in_channels, out_channels, kernel_size=3, padding=1),
                nn.BatchNorm2d(out_channels),
                nn.PReLU(),
                nn.Dropout(dropout_p),
                nn.Conv2d(out_channels, out_channels, kernel_size=3, padding=1),
                nn.BatchNorm2d(out_channels),
                nn.PReLU()
            )
        else:
            self.conv_conv = nn.Sequential(
                nn.Conv3d(in_channels, out_channels, kernel_size=3, padding=1),
                nn.BatchNorm3d(out_channels),
                nn.PReLU(),
                nn.Dropout(dropout_p),
                nn.Conv3d(out_channels, out_channels, kernel_size=3, padding=1),
                nn.BatchNorm3d(out_channels),
                nn.PReLU()
            )

    def forward(self, x):
        output = self.conv_conv(x)
        return output 

class DownBlock(nn.Module):
    """a convolutional block followed by downsampling"""
    def __init__(self,in_channels, out_channels, 
                dim = 2, dropout_p = 0.0, downsample = True):
        super(DownBlock, self).__init__()
        self.downsample = downsample 
        self.dim = dim
        self.conv = ConvBlockND(in_channels, out_channels, dim, dropout_p)
        if(downsample):
            if(self.dim == 2):
                self.down_layer = nn.MaxPool2d(kernel_size = 2, stride = 2)
            else:
                self.down_layer = nn.MaxPool3d(kernel_size = 2, stride = 2)
    
    def forward(self, x):
        x_shape = list(x.shape)
        if(self.dim == 2 and len(x_shape) == 5):
            [N, C, D, H, W] = x_shape
            new_shape = [N*D, C, H, W]
            x = torch.transpose(x, 1, 2)
            x = torch.reshape(x, new_shape)
        output = self.conv(x)
        if(self.downsample):
            output_d = self.down_layer(output)
        else:
            output_d = None 
        if(self.dim == 2 and len(x_shape) == 5):
            new_shape = [N, D] + list(output.shape)[1:]
            output = torch.reshape(output, new_shape)
            output = torch.transpose(output, 1, 2)
            if(self.downsample):
                new_shape = [N, D] + list(output_d.shape)[1:]
                output_d = torch.reshape(output_d, new_shape)
                output_d = torch.transpose(output_d, 1, 2)

        return output, output_d

class UpBlock(nn.Module):
    """Upsampling followed by ConConvBlockNDvBlock"""
    def __init__(self, in_channels1, in_channels2, out_channels, 
                 dim = 2, dropout_p = 0.0, bilinear=True):
        super(UpBlock, self).__init__()
        self.bilinear = bilinear
        self.dim = dim
        if bilinear:
            if(dim == 2):
                self.up = nn.Sequential(
                    nn.Conv2d(in_channels1, in_channels2, kernel_size = 1),
                    nn.Upsample(scale_factor=2, mode='bilinear', align_corners=True))
            else:
                self.up = nn.Sequential(
                    nn.Conv3d(in_channels1, in_channels2, kernel_size = 1),
                    nn.Upsample(scale_factor=2, mode='trilinear', align_corners=True))
        else:
            if(dim == 2):
                self.up = nn.ConvTranspose2d(in_channels1, in_channels2, kernel_size=2, stride=2)
            else:
                self.up = nn.ConvTranspose3d(in_channels1, in_channels2, kernel_size=2, stride=2)
            
        self.conv = ConvBlockND(in_channels2 * 2, out_channels, dim, dropout_p)

    def forward(self, x1, x2):
        x1_shape = list(x1.shape)
        x2_shape = list(x2.shape)
        if(self.dim == 2 and len(x1_shape) == 5):
            [N, C, D, H, W] = x1_shape
            new_shape = [N*D, C, H, W]
            x1 = torch.transpose(x1, 1, 2)
            x1 = torch.reshape(x1, new_shape)
            [N, C, D, H, W] = x2_shape
            new_shape = [N*D, C, H, W]
            x2 = torch.transpose(x2, 1, 2)
            x2 = torch.reshape(x2, new_shape)

        x1 = self.up(x1)
        output = torch.cat([x2, x1], dim=1)
        output = self.conv(output)
        if(self.dim == 2 and len(x1_shape) == 5):
            new_shape = [N, D] + list(output.shape)[1:]
            output = torch.reshape(output, new_shape)
            output = torch.transpose(output, 1, 2)
        return output  

class UNet2D5(nn.Module):
    def __init__(self, params):
        super(UNet2D5, self).__init__()
        self.params    = params
        self.in_chns   = self.params['in_chns']
        self.ft_chns   = self.params['feature_chns']
        self.dims      = self.params['conv_dims']
        self.n_class   = self.params['class_num']
        self.bilinear  = self.params['bilinear']
        self.dropout   = self.params['dropout']
        assert(len(self.ft_chns) == 5)

        self.block0 = DownBlock(self.in_chns, self.ft_chns[0], self.dims[0], self.dropout[0], True)
        self.block1 = DownBlock(self.ft_chns[0], self.ft_chns[1], self.dims[1], self.dropout[1], True)
        self.block2 = DownBlock(self.ft_chns[1], self.ft_chns[2], self.dims[2], self.dropout[2], True)
        self.block3 = DownBlock(self.ft_chns[2], self.ft_chns[3], self.dims[3], self.dropout[3], True)
        self.block4 = DownBlock(self.ft_chns[3], self.ft_chns[4], self.dims[4], self.dropout[4], False)
        self.up1 = UpBlock(self.ft_chns[4], self.ft_chns[3], self.ft_chns[3], 
                    self.dims[3], dropout_p = self.dropout[3], bilinear = self.bilinear) 
        self.up2 = UpBlock(self.ft_chns[3], self.ft_chns[2], self.ft_chns[2], 
                    self.dims[2], dropout_p = self.dropout[2], bilinear = self.bilinear) 
        self.up3 = UpBlock(self.ft_chns[2], self.ft_chns[1], self.ft_chns[1], 
                    self.dims[1], dropout_p = self.dropout[1], bilinear = self.bilinear) 
        self.up4 = UpBlock(self.ft_chns[1], self.ft_chns[0], self.ft_chns[0], 
                    self.dims[0], dropout_p = self.dropout[0], bilinear = self.bilinear) 
    
        self.out_conv = nn.Conv3d(self.ft_chns[0], self.n_class, 
            kernel_size = (1, 3, 3), padding = (0, 1, 1))

    def forward(self, x):
        x0, x0_d = self.block0(x)
        x1, x1_d = self.block1(x0_d)
        x2, x2_d = self.block2(x1_d)
        x3, x3_d = self.block3(x2_d)
        x4, x4_d = self.block4(x3_d)
        
        x = self.up1(x4, x3)
        x = self.up2(x, x2)
        x = self.up3(x, x1)
        x = self.up4(x, x0)
        output = self.out_conv(x)
        return output


if __name__ == "__main__":
    params = {'in_chns':4,
              'feature_chns':[2, 8, 32, 48, 64],
              'conv_dims': [2, 2, 3, 3, 3],
              'dropout':  [0, 0, 0.3, 0.4, 0.5],
              'class_num': 2,
              'bilinear': False}
    Net = UNet2D5(params)
    Net = Net.double()

    x  = np.random.rand(4, 4, 32, 96, 96)
    xt = torch.from_numpy(x)
    xt = torch.tensor(xt)
    
    y = Net(xt)
    print(len(y.size()))
    y = y.detach().numpy()
    print(y.shape)



# -*- coding: utf-8 -*-
from __future__ import print_function, division
from pymic.net.cls.torch_pretrained_net import *

TorchClsNetDict = {
    'resnet18':   ResNet18,
    'vgg16':      VGG16,
    'mobilenetv2':MobileNetV2
}



# -*- coding: utf-8 -*-
from __future__ import print_function, division

import torch
import json
import math
import random
import numpy as np
from scipy import ndimage
from pymic.transform.abstract_transform import AbstractTransform
from pymic.util.image_process import *


class GammaCorrection(AbstractTransform):
    """
    apply random gamma correction to each channel
    """
    def __init__(self, params):
        """
        (gamma_min, gamma_max) specify the range of gamma
        """
        super(GammaCorrection, self).__init__(params)
        self.channels =  params['GammaCorrection_channels'.lower()]
        self.gamma_min = params['GammaCorrection_gamma_min'.lower()]
        self.gamma_max = params['GammaCorrection_gamma_max'.lower()]
        self.prob      = params.get('GammaCorrection_probability'.lower(), 0.5)
        self.inverse = params.get('GammaCorrection_inverse'.lower(), False)
    
    def __call__(self, sample):
        if(np.random.uniform() > self.prob):
            return sample
        image= sample['image']
        for chn in self.channels:
            gamma_c = random.random() * (self.gamma_max - self.gamma_min) + self.gamma_min
            img_c = image[chn]
            v_min = img_c.min()
            v_max = img_c.max()
            img_c = (img_c - v_min)/(v_max - v_min)
            img_c = np.power(img_c, gamma_c)*(v_max - v_min) + v_min
            image[chn] = img_c

        sample['image'] = image
        return sample

class GaussianNoise(AbstractTransform):
    """
    apply random gamma correction to each channel
    """
    def __init__(self, params):
        """
        (gamma_min, gamma_max) specify the range of gamma
        """
        super(GaussianNoise, self).__init__(params)
        self.channels = params['GaussianNoise_channels'.lower()]
        self.mean     = params['GaussianNoise_mean'.lower()]
        self.std      = params['GaussianNoise_std'.lower()]
        self.prob     = params.get('GaussianNoise_probability'.lower(), 0.5)
        self.inverse  = params.get('GaussianNoise_inverse'.lower(), False)
    
    def __call__(self, sample):
        if(np.random.uniform() > self.prob):
            return sample
        image= sample['image']
        for chn in self.channels:
            img_c = image[chn]
            noise = np.random.normal(self.mean, self.std, img_c.shape)
            image[chn] = img_c + noise

        sample['image'] = image
        return sample



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function

import numpy as np
import SimpleITK as sitk
from scipy import ndimage

def get_ND_bounding_box(volume, margin = None):
    """
    get the bounding box of nonzero region in an ND volume
    """
    input_shape = volume.shape
    if(margin is None):
        margin = [0] * len(input_shape)
    assert(len(input_shape) == len(margin))
    indxes = np.nonzero(volume)
    idx_min = []
    idx_max = []
    for i in range(len(input_shape)):
        idx_min.append(int(indxes[i].min()))
        idx_max.append(int(indxes[i].max()) + 1)

    for i in range(len(input_shape)):
        idx_min[i] = max(idx_min[i] - margin[i], 0)
        idx_max[i] = min(idx_max[i] + margin[i], input_shape[i])
    return idx_min, idx_max

def crop_ND_volume_with_bounding_box(volume, min_idx, max_idx):
    """
    crop/extract a subregion form an nd image.
    """
    dim = len(volume.shape)
    assert(dim >= 2 and dim <= 5)
    assert(max_idx[0] - min_idx[0] <= volume.shape[0])
    if(dim == 2):
        output = volume[min_idx[0]:max_idx[0], min_idx[1]:max_idx[1]]
    elif(dim == 3):
        output = volume[min_idx[0]:max_idx[0], min_idx[1]:max_idx[1], min_idx[2]:max_idx[2]]
    elif(dim == 4):
        output = volume[min_idx[0]:max_idx[0], min_idx[1]:max_idx[1], min_idx[2]:max_idx[2], min_idx[3]:max_idx[3]]
    elif(dim == 5):
        output = volume[min_idx[0]:max_idx[0], min_idx[1]:max_idx[1], min_idx[2]:max_idx[2], min_idx[3]:max_idx[3], min_idx[4]:max_idx[4]]
    else:
        raise ValueError("the dimension number shoud be 2 to 5")
    return output

def set_ND_volume_roi_with_bounding_box_range(volume, bb_min, bb_max, sub_volume, addition = True):
    """
    set a subregion to an nd image. if addition is True, the original volume is added by the subregion.
    """
    dim = len(bb_min)
    out = volume
    if(dim == 2):
        if(addition):
            out[bb_min[0]:bb_max[0], bb_min[1]:bb_max[1]] += sub_volume
        else:
            out[bb_min[0]:bb_max[0], bb_min[1]:bb_max[1]] = sub_volume
    elif(dim == 3):
        if(addition):
            out[bb_min[0]:bb_max[0], bb_min[1]:bb_max[1], bb_min[2]:bb_max[2]] += sub_volume
        else:
            out[bb_min[0]:bb_max[0], bb_min[1]:bb_max[1], bb_min[2]:bb_max[2]] = sub_volume
    elif(dim == 4):
        if(addition):
            out[bb_min[0]:bb_max[0], bb_min[1]:bb_max[1], bb_min[2]:bb_max[2], bb_min[3]:bb_max[3]]  += sub_volume
        else:
            out[bb_min[0]:bb_max[0], bb_min[1]:bb_max[1], bb_min[2]:bb_max[2], bb_min[3]:bb_max[3]] = sub_volume
    elif(dim == 5):
        if(addition):
            out[bb_min[0]:bb_max[0], bb_min[1]:bb_max[1], bb_min[2]:bb_max[2], bb_min[3]:bb_max[3], bb_min[4]:bb_max[4]] += sub_volume
        else:
            out[bb_min[0]:bb_max[0], bb_min[1]:bb_max[1], bb_min[2]:bb_max[2], bb_min[3]:bb_max[3], bb_min[4]:bb_max[4]] = sub_volume
    else:
        raise ValueError("array dimension should be 2 to 5")
    return out

def crop_and_pad_ND_array_to_desired_shape(image, out_shape, pad_mod):
    in_shape   = image.shape 
    dim        = len(in_shape)
    crop_shape = [min(out_shape[i], in_shape[i])  for i in range(dim)]
    mgnc = [max(0, in_shape[i] - crop_shape[i]) for i in range(dim)]
    if(max(mgnc) == 0):
        image_crp = image
    else:
        ml   = [int(mgnc[i]/2)  for i in range(dim)]
        mr   = [mgnc[i] - ml[i] for i in range(dim)] 
        if(dim == 2):
            image_crp = image[ml[0]:(in_shape[0] - mr[0]), ml[1]:(in_shape[1] - mr[1])]
        elif(dim == 3):
            image_crp = image[ml[0]:(in_shape[0] - mr[0]), ml[1]:(in_shape[1] - mr[1]), ml[2]:(in_shape[2] - mr[2])]
        elif(dim == 4):
            image_crp = image[ml[0]:(in_shape[0] - mr[0]), ml[1]:(in_shape[1] - mr[1]), ml[2]:(in_shape[2] - mr[2]), ml[3]:(in_shape[3] - mr[3])]
        elif(dim == 5):
            image_crp = image[ml[0]:(in_shape[0] - mr[0]), ml[1]:(in_shape[1] - mr[1]), ml[2]:(in_shape[2] - mr[2]), ml[3]:(in_shape[3] - mr[3]), ml[4]:(in_shape[4] - mr[4])]
        else:
            raise ValueError("array dimension should be 2 to 5")

    mgnp = [out_shape[i] - crop_shape[i] for i in range(dim)]
    if(max(mgnp) == 0):
        image_pad = image_crp
    else:
        ml   = [int(mgnp[i]/2)  for i in range(dim)]
        mr   = [mgnp[i] - ml[i] for i in range(dim)] 
        pad  = [(ml[i], mr[i])  for i in range(dim)]
        pad  = tuple(pad)
        image_pad = np.pad(image_crp, pad, pad_mod) 
        
    return image_pad

def get_largest_k_components(image, k = 1):
    """
    get the largest K components from 2D or 3D binary image
    image: nd array
    """
    dim = len(image.shape)
    if(image.sum() == 0 ):
        print('the largest component is null')
        return image
    if(dim < 2 or dim > 3):
        raise ValueError("the dimension number should be 2 or 3")
    s = ndimage.generate_binary_structure(dim,1)      
    labeled_array, numpatches = ndimage.label(image, s)
    sizes = ndimage.sum(image, labeled_array, range(1, numpatches + 1))
    sizes_sort = sorted(sizes, reverse = True)
    kmin = min(k, numpatches)
    output = np.zeros_like(image)
    for i in range(kmin):
        labeli = np.where(sizes == sizes_sort[i])[0] + 1
        output = output + np.asarray(labeled_array == labeli, np.uint8)
    return  output

def get_euclidean_distance(image, dim = 3, spacing = [1.0, 1.0, 1.0]):
    """
    get euclidean distance transform of 2D or 3D binary images
    the output distance map is unsigned
    """
    img_shape = image.shape
    input_dim = len(img_shape)
    if(input_dim != 3):
        raise ValueError("Not implemented for {0:}D image".format(input_dim))
    if(dim == 2):
        raise ValueError("Not implemented for {0:}D image".format(input_dim))
        # dis_map = np.ones_like(image, np.float32)
        # for d in range(img_shape[0]):
        #     if(image[d].sum() > 0):
        #         dis_d = ndimage.morphology.distance_transform_edt(image[d])
        #         dis_map[d] = dis_d/dis_d.max()
    elif(dim == 3):
        fg_dis_map = ndimage.morphology.distance_transform_edt(image > 0.5)
        bg_dis_map = ndimage.morphology.distance_transform_edt(image <= 0.5)
        dis_map = bg_dis_map - fg_dis_map
    else:
        raise ValueError("Not implemented for {0:}D distance".format(dim))
    return dis_map

def convert_label(label, source_list, target_list):
    """
    convert a label map based a source list and a target list of labels
    label: nd array 
    source_list: a list of labels that will be converted, e.g. [0, 1, 2, 4]
    target_list: a list of target labels, e.g. [0, 1, 2, 3]
    """
    assert(len(source_list) == len(target_list))
    label_converted = np.zeros_like(label)
    for i in range(len(source_list)):
        label_temp = np.asarray(label == source_list[i], label.dtype)
        label_temp = label_temp * target_list[i]
        label_converted = label_converted + label_temp
    return label_converted

def resample_sitk_image_to_given_spacing(image, spacing, order):
    """
    image: an sitk image object
    spacing: 3D tuple / list for spacing along x, y, z direction
    order: order for interpolation
    """
    spacing0 = image.GetSpacing()
    data = sitk.GetArrayFromImage(image)
    zoom = [spacing0[i] / spacing[i] for i in range(3)]
    zoom = [zoom[2], zoom[0], zoom[1]]
    data = ndimage.interpolation.zoom(data, zoom, order = order)
    out_img = sitk.GetImageFromArray(data)
    out_img.SetSpacing(spacing)
    out_img.SetDirection(image.GetDirection())
    return out_img



# -*- coding: utf-8 -*-
from __future__ import print_function, division

import torch
import json
import math
import random
import numpy as np
from scipy import ndimage
from pymic.transform.abstract_transform import AbstractTransform
from pymic.util.image_process import *

class ReduceLabelDim(AbstractTransform):
    """
    remove the first dimension of label tensor
    """
    def __init__(self, params):
        super(ReduceLabelDim, self).__init__(params)
        self.inverse = params.get('ReduceLabelDim_inverse'.lower(), False)
    
    def __call__(self, sample):
        label = sample['label']
        label_converted = label[0]
        sample['label'] = label_converted
        return sample

class LabelConvert(AbstractTransform):
    """ Convert a list of labels to another list"""
    def __init__(self, params):
        """
        source_list (tuple/list): A list of labels to be converted
        target_list (tuple/list): The target label list
        """
        super(LabelConvert, self).__init__(params)
        self.source_list = params['LabelConvert_source_list'.lower()]
        self.target_list = params['LabelConvert_target_list'.lower()]
        self.inverse = params.get('LabelConvert_inverse'.lower(), False)
        assert(len(self.source_list) == len(self.target_list))
    
    def __call__(self, sample):
        label = sample['label']
        label_converted = convert_label(label, self.source_list, self.target_list)
        sample['label'] = label_converted
        return sample

class LabelConvertNonzero(AbstractTransform):
    """ Convert label into binary (nonzero as 1)"""
    def __init__(self, params):
        super(LabelConvertNonzero, self).__init__(params)
        self.inverse = params.get('LabelConvertNonzero_inverse'.lower(), False)
    
    def __call__(self, sample):
        label = sample['label']
        label_converted = np.asarray(label > 0, np.uint8)
        sample['label'] = label_converted
        return sample

class LabelToProbability(AbstractTransform):
    """Convert one-channel label map to one-hot multi-channel probability map"""
    def __init__(self, params): 
        """
        class_num (int): the class number in the label map
        """
        super(LabelToProbability, self).__init__(params)
        self.class_num = params['LabelToProbability_class_num'.lower()]
        self.inverse   = params.get('LabelToProbability_inverse'.lower(), False)

    def __call__(self, sample):
        if(self.task == 'segmentation'):
            label = sample['label'][0] # sample['label'] is (1, h, w)
            label_prob = np.zeros((self.class_num, *label.shape), dtype = np.float32)
            for i in range(self.class_num):
                label_prob[i] = label == i*np.ones_like(label)
            sample['label_prob'] = label_prob
        elif(self.task == 'classification'):
            label_idx = sample['label']
            label_prob = np.zeros((self.class_num,), np.float32)
            label_prob[label_idx] = 1.0
            sample['label_prob'] = label_prob 
        return sample


class PartialLabelToProbability(AbstractTransform):
    """Convert one-channel label map to one-hot multi-channel probability map
    Note that the label map represents partial labels.
    For segmentation tasks only. 
    0: background
    1 to C-1: foreground (C-classes)
    C: unknown label. 
    the output consists of:
    label_prob: one-hot probability map
    pixel_weight: weigh of pixels, 0 if the label is unknown
    """
    def __init__(self, params): 
        """
        class_num (int): the class number in the label map
        """
        super(PartialLabelToProbability, self).__init__(params)
        self.class_num = params['PartialLabelToProbability_class_num'.lower()]
        self.inverse   = params.get('PartialLabelToProbability_inverse'.lower(), False)
    
    def __call__(self, sample):
        label = sample['label'][0]
        assert(label.max() <= self.class_num)
        label_prob = np.zeros((self.class_num, *label.shape), dtype = np.float32)
        for i in range(self.class_num):
            label_prob[i] = label == i*np.ones_like(label)
        sample['label_prob'] = label_prob
        sample['pixel_weight'] = 1.0 - np.asarray([label == self.class_num], np.float32)

        # # for gated CRF loss
        # scribble = label - 1
        # scribble[label == 0] = 255
        # sample['scribbles'] = scribble
        return sample







# -*- coding: utf-8 -*-
from __future__ import print_function, division

class AbstractTransform(object):
    def __init__(self, params):
        self.task = params['Task'.lower()]

    def __call__(self, sample):
        return sample

    def inverse_transform_for_prediction(self, sample):
        raise(ValueError("not implemented"))



# -*- coding: utf-8 -*-
from __future__ import print_function, division
import torch
import numpy as np 

def keyword_match(a,b):
    return a.lower() == b.lower()

def get_one_hot_seg(label, class_num):
    """
    convert a segmentation label to one-hot
    label: a tensor with a shape of [N, 1, D, H, W] or [N, 1, H, W]
    class_num: class number. 
    output: an one-hot tensor with a shape of [N, C, D, H, W] or [N, C, H, W]
    """
    size = list(label.size())
    if(size[1] != 1):
        raise ValueError("The channel should be 1, \
            rather than {0:} before one-hot encoding".format(size[1]))
    label = label.view(-1)
    ones  = torch.sparse.torch.eye(class_num).to(label.device)
    one_hot = ones.index_select(0, label)
    size.append(class_num)
    one_hot = one_hot.view(*size)
    one_hot = torch.transpose(one_hot, 1, -1)
    one_hot = torch.squeeze(one_hot, -1)
    return one_hot


# -*- coding: utf-8 -*-
from __future__ import print_function, division

import torch
import torch.nn as nn

class CombinedLoss(nn.Module):
    def __init__(self, params, loss_dict):
        super(CombinedLoss, self).__init__()
        loss_names  = params['loss_type']
        self.loss_weight = params['loss_weight']
        assert (len(loss_names) == len(self.loss_weight))
        self.loss_list = []
        for loss_name in loss_names:
            if(loss_name in loss_dict):
                one_loss = loss_dict[loss_name](params)
                self.loss_list.append(one_loss)
            else:
                raise ValueError("{0:} is not defined, or has not been added to the \
                    loss dictionary".format(loss_name))

    def forward(self, loss_input_dict):
        loss_value = 0.0
        for i in range(len(self.loss_list)):
            loss_value += self.loss_weight[i]*self.loss_list[i](loss_input_dict)
        return loss_value



# -*- coding: utf-8 -*-
from __future__ import print_function, division

import torch
import torch.nn as nn

class ConvolutionLayer(nn.Module):
    """
    A compose layer with the following components:
    convolution -> (batch_norm / layer_norm / group_norm / instance_norm) -> activation -> (dropout)
    batch norm and dropout are optional
    """
    def __init__(self, in_channels, out_channels, kernel_size, dim = 3,
            stride = 1, padding = 0, dilation = 1, conv_group = 1, bias = True, 
            norm_type = 'batch_norm', norm_group = 1, acti_func = None):
        super(ConvolutionLayer, self).__init__()
        self.n_in_chns  = in_channels
        self.n_out_chns = out_channels
        self.norm_type  = norm_type
        self.norm_group = norm_group
        self.acti_func  = acti_func

        assert(dim == 2 or dim == 3)
        if(dim == 2):
            self.conv = nn.Conv2d(in_channels, out_channels,
                kernel_size, stride, padding, dilation, conv_group, bias)
            if(self.norm_type == 'batch_norm'):
                self.bn = nn.BatchNorm2d(out_channels)
            elif(self.norm_type == 'group_norm'):
                self.bn = nn.GroupNorm(self.norm_group, out_channels)
            elif(self.norm_type is not None):
                raise ValueError("unsupported normalization method {0:}".format(norm_type))
        else:        
            self.conv = nn.Conv3d(in_channels, out_channels,
                kernel_size, stride, padding, dilation, conv_group, bias)
            if(self.norm_type == 'batch_norm'):
                self.bn = nn.BatchNorm3d(out_channels)
            elif(self.norm_type == 'group_norm'):
                self.bn = nn.GroupNorm(self.norm_group, out_channels)
            elif(self.norm_type is not None):
                raise ValueError("unsupported normalization method {0:}".format(norm_type))

    def forward(self, x):
        f = self.conv(x)
        if(self.norm_type is not None):
            f = self.bn(f)
        if(self.acti_func is not None):
            f = self.acti_func(f)
        return f

class DepthSeperableConvolutionLayer(nn.Module):
    """
    A compose layer with the following components:
    convolution -> (batch_norm) -> activation -> (dropout)
    batch norm and dropout are optional
    """
    def __init__(self, in_channels, out_channels, kernel_size, dim = 3,
            stride = 1, padding = 0, dilation =1, conv_group = 1, bias = True, 
            norm_type = 'batch_norm', norm_group = 1, acti_func = None):
        super(DepthSeperableConvolutionLayer, self).__init__()
        self.n_in_chns  = in_channels
        self.n_out_chns = out_channels
        self.norm_type  = norm_type
        self.norm_group = norm_group
        self.acti_func  = acti_func

        assert(dim == 2 or dim == 3)
        if(dim == 2):
            self.conv1x1 = nn.Conv2d(in_channels, out_channels,
                kernel_size = 1, stride = stride, padding = 0, dilation = dilation, groups = conv_group, bias = bias)
            self.conv = nn.Conv2d(out_channels, out_channels,
                kernel_size, stride, padding, dilation, groups = out_channels, bias = bias)
            if(self.norm_type == 'batch_norm'):
                self.bn = nn.BatchNorm2d(out_channels)
            elif(self.norm_type == 'group_norm'):
                self.bn = nn.GroupNorm(self.norm_group, out_channels)
            elif(self.norm_type is not None):
                raise ValueError("unsupported normalization method {0:}".format(norm_type))
        else:     
            self.conv1x1 = nn.Conv3d(in_channels, out_channels,
                kernel_size = 1, stride = stride, padding = 0, dilation = dilation, groups = conv_group, bias = bias)   
            self.conv = nn.Conv3d(out_channels, out_channels,
                kernel_size, stride, padding, dilation, groups = out_channels, bias = bias)
            if(self.norm_type == 'batch_norm'):
                self.bn = nn.BatchNorm3d(out_channels)
            elif(self.norm_type == 'group_norm'):
                self.bn = nn.GroupNorm(self.norm_group, out_channels)
            elif(self.norm_type is not None):
                raise ValueError("unsupported normalization method {0:}".format(norm_type))

    def forward(self, x):
        f = self.conv1x1(x)
        f = self.conv(f)
        if(self.norm_type is not None):
            f = self.bn(f)
        if(self.acti_func is not None):
            f = self.acti_func(f)
        return f

class ConvolutionSepAll3DLayer(nn.Module):
    """
    A compose layer with the following components:
    convolution -> (batch_norm) -> activation -> (dropout)
    batch norm and dropout are optional
    """
    def __init__(self, in_channels, out_channels, kernel_size, dim = 3,
            stride = 1, padding = 0, dilation =1, groups = 1, bias = True, 
            batch_norm = True, acti_func = None):
        super(ConvolutionSepAll3DLayer, self).__init__()
        self.n_in_chns  = in_channels
        self.n_out_chns = out_channels
        self.batch_norm = batch_norm
        self.acti_func  = acti_func

        assert(dim == 3)
        chn = min(in_channels, out_channels)

        self.conv_intra_plane1 = nn.Conv2d(chn, chn,
            kernel_size, stride, padding, dilation, chn, bias)

        self.conv_intra_plane2 = nn.Conv2d(chn, chn,
            kernel_size, stride, padding, dilation, chn, bias)

        self.conv_intra_plane3 = nn.Conv2d(chn, chn,
            kernel_size, stride, padding, dilation, chn, bias)

        self.conv_space_wise  = nn.Conv2d(in_channels, out_channels,
            1, stride, 0, dilation, 1, bias)
        
        if(self.batch_norm):
                self.bn = nn.BatchNorm3d(out_channels)

    def forward(self, x):
        in_shape = list(x.shape)
        assert(len(in_shape) == 5)
        [B, C, D, H, W] = in_shape
        f0 = x.permute(0, 2, 1, 3, 4)  #[B, D, C, H, W] 
        f0 = f0.contiguous().view([B*D, C, H, W])

        Cc = min(self.n_in_chns, self.n_out_chns)
        Co = self.n_out_chns
        if(self.n_in_chns > self.n_out_chns):
            f0 = self.conv_space_wise(f0)  #[B*D, Cc, H, W]  
   
        f1 = self.conv_intra_plane1(f0)
        f2 = f1.contiguous().view([B, D, Cc, H, W])
        f2 = f2.permute(0, 3, 2, 1, 4) #[B, H, Cc, D, W]
        f2 = f2.contiguous().view([B*H, Cc, D, W])
        f2 = self.conv_intra_plane2(f2)
        f3 = f2.contiguous().view([B, H, Cc, D, W])
        f3 = f3.permute(0, 4, 2, 3, 1) #[B, W, Cc, D, H]
        f3 = f3.contiguous().view([B*W, Cc, D, H])
        f3 = self.conv_intra_plane3(f3)
        if(self.n_in_chns <= self.n_out_chns):
            f3 = self.conv_space_wise(f3)  #[B*W, Co, D, H] 

        f3 = f3.contiguous().view([B, W, Co, D, H])
        f3 = f3.permute([0, 2, 3, 4, 1]) #[B, Co, D, H, W]

        if(self.batch_norm):
            f3 = self.bn(f3)
        if(self.acti_func is not None):
            f3 = self.acti_func(f3)
        return f3



# -*- coding: utf-8 -*-
from __future__ import print_function, division

import torch
import torch.nn as nn
import numpy as np

def get_soft_label(input_tensor, num_class, data_type = 'float'):
    """
        convert a label tensor to one-hot soft label 
        input_tensor: tensor with shae [B, 1]
        output_tensor: shape [B, num_class]
    """
    tensor_list = []
    for i in range(num_class):
        temp_prob = input_tensor == i*torch.ones_like(input_tensor)
        tensor_list.append(temp_prob)
    output_tensor = torch.cat(tensor_list, dim = 1)
    if(data_type == 'float'):
        output_tensor = output_tensor.float()
    elif(data_type == 'double'):
        output_tensor = output_tensor.double()
    else:
        raise ValueError("data type can only be float and double: {0:}".format(data_type))

    return output_tensor


# -*- coding: utf-8 -*-
from __future__ import print_function, division

import torch
from torch import optim
from torch.optim import lr_scheduler
from pymic.util.general import keyword_match

def get_optimizer(name, net_params, optim_params):
    lr = optim_params['learning_rate']
    momentum = optim_params['momentum']
    weight_decay = optim_params['weight_decay']
    if(keyword_match(name, "SGD")):
        return optim.SGD(net_params, lr, 
            momentum = momentum, weight_decay = weight_decay)
    elif(keyword_match(name, "Adam")):
        return optim.Adam(net_params, lr, weight_decay = weight_decay)
    elif(keyword_match(name, "SparseAdam")):
        return optim.SparseAdam(net_params, lr)
    elif(keyword_match(name, "Adadelta")):
        return optim.Adadelta(net_params, lr, weight_decay = weight_decay)
    elif(keyword_match(name, "Adagrad")): 
        return optim.Adagrad(net_params, lr, weight_decay = weight_decay)
    elif(keyword_match(name, "Adamax")): 
        return optim.Adamax(net_params, lr, weight_decay = weight_decay)
    elif(keyword_match(name, "ASGD")): 
        return optim.ASGD(net_params, lr, weight_decay = weight_decay)
    elif(keyword_match(name, "LBFGS")): 
        return optim.LBFGS(net_params, lr)
    elif(keyword_match(name, "RMSprop")): 
        return optim.RMSprop(net_params, lr, momentum = momentum,
            weight_decay = weight_decay)
    elif(keyword_match(name, "Rprop")): 
        return optim.Rprop(net_params, lr)
    else:
        raise ValueError("unsupported optimizer {0:}".format(name))


def get_lr_scheduler(optimizer, sched_params):
    name = sched_params["lr_scheduler"]
    if(name is None):
        return None
    lr_gamma = sched_params["lr_gamma"]
    if(keyword_match(name, "ReduceLROnPlateau")):
        patience_it = sched_params["ReduceLROnPlateau_patience".lower()]
        val_it = sched_params["iter_valid"]
        patience = patience_it / val_it
        scheduler = lr_scheduler.ReduceLROnPlateau(optimizer,
            mode = "max", factor=lr_gamma, patience = patience)
    elif(keyword_match(name, "MultiStepLR")):
        lr_milestones = sched_params["lr_milestones"]
        last_iter     = sched_params["last_iter"]
        scheduler = lr_scheduler.MultiStepLR(optimizer,
                    lr_milestones, lr_gamma, last_iter)
    else:
        raise ValueError("unsupported lr scheduler {0:}".format(name))
    return scheduler


# -*- coding: utf-8 -*-
from __future__ import print_function, division

import torch
import json
import math
import random
import numpy as np
from scipy import ndimage
from pymic.transform.abstract_transform import AbstractTransform
from pymic.util.image_process import *


class Pad(AbstractTransform):
    """
    Pad the image (shape [C, D, H, W] or [C, H, W]) to an new spatial shape, 
    the real output size will be max(image_size, output_size)
    """
    def __init__(self, params):
        """
        :param output_size: (tuple/list) the size along each spatial axis. 
        :param ceil_mode: (bool) if true, the real output size is integer multiples of output_size.
        """
        super(Pad, self).__init__(params)
        self.output_size = params['Pad_output_size'.lower()]
        self.ceil_mode   = params['Pad_ceil_mode'.lower()]
        self.inverse = params.get('Pad_inverse'.lower(), True)

    def __call__(self, sample):
        image = sample['image']
        input_shape = image.shape
        input_dim = len(input_shape) - 1
        assert(len(self.output_size) == input_dim)
        if(self.ceil_mode):
            multiple = [int(math.ceil(float(input_shape[1+i])/self.output_size[i]))\
                for i in range(input_dim)]
            output_size = [multiple[i] * self.output_size[i] \
                for i in range(input_dim)]
        else:
            output_size = self.output_size
        margin = [max(0, output_size[i] - input_shape[1+i]) \
            for i in range(input_dim)]

        margin_lower = [int(margin[i] / 2) for i in range(input_dim)]
        margin_upper = [margin[i] - margin_lower[i] for i in range(input_dim)]
        sample['Pad_Param'] = json.dumps((margin_lower, margin_upper))

        pad = [(margin_lower[i], margin_upper[i]) for  i in range(input_dim)]
        pad = tuple([(0, 0)] + pad)
        image_t = np.pad(image, pad, 'reflect') if(max(margin) > 0) else image

        sample['image'] = image_t
        
        if('label' in sample and self.task == 'segmentation'):
            label = sample['label']
            label = np.pad(label, pad, 'reflect') if(max(margin) > 0) else label
            sample['label'] = label
        if('pixel_weight' in sample and self.task == 'segmentation'):
            weight = sample['pixel_weight']
            weight = np.pad(weight, pad, 'reflect') if(max(margin) > 0) else weight
            sample['pixel_weight'] = weight
        return sample
    
    def inverse_transform_for_prediction(self, sample):
        ''' crop sample['predict'] (5D or 4D) to the original spatial shape.
        origin_shape is a 4D or 3D vector as saved in __call__().'''
        # raise ValueError("not implemented")
        if(isinstance(sample['Pad_Param'], list) or isinstance(sample['Pad_Param'], tuple)):
            params = json.loads(sample['Pad_Param'][0]) 
        else:
            params = json.loads(sample['Pad_Param']) 
        margin_lower = params[0]
        margin_upper = params[1]
        predict = sample['predict']
        if(isinstance(predict, tuple) or isinstance(predict, list)):
            output_predict = []
            for predict_i in predict:
                predict_shape = predict_i.shape
                crop_min = [0, 0] + margin_lower
                crop_max = [predict_shape[2:][i] - margin_upper[i] \
                    for i in range(len(margin_lower))]
                crop_max = list(predict_shape[:2]) + crop_max
                crop_predict = crop_ND_volume_with_bounding_box(predict_i, crop_min, crop_max)
                output_predict.append(crop_predict)
        else:
            predict_shape = predict.shape
            crop_min = [0, 0] + margin_lower
            crop_max = [predict_shape[2:][i] - margin_upper[i] \
                for i in range(len(margin_lower))]
            crop_max = list(predict_shape[:2]) + crop_max

            output_predict = crop_ND_volume_with_bounding_box(predict, crop_min, crop_max)
        sample['predict'] = output_predict
        return sample


"""
3D implementation of Spacial channel Squeeze and Excitation 
   <https://arxiv.org/abs/1709.01507>`_
`Spatial Squeeze and Excitation <https://arxiv.org/abs/1803.02579>`
*****************************

oringinal file: https://github.com/maodong2056/squeeze_and_excitation/blob/master/squeeze_and_excitation/squeeze_and_excitation.py
"""

from enum import Enum
import torch
import torch.nn as nn
import torch.nn.functional as F

class ChannelSELayer3D(nn.Module):
    """
    3D implementation of Squeeze-and-Excitation (SE) block described in:
        *Hu et al., Squeeze-and-Excitation Networks, arXiv:1709.01507*
    """
    def __init__(self, num_channels, reduction_ratio=2):

        """
        :param num_channels: No of input channels
        :param reduction_ratio: By how much should the num_channels should be reduced
        """
        super(ChannelSELayer3D, self).__init__()
        num_channels_reduced = num_channels // reduction_ratio
        self.reduction_ratio = reduction_ratio
        self.fc1 = nn.Linear(num_channels, num_channels_reduced, bias=True)
        self.fc2 = nn.Linear(num_channels_reduced, num_channels, bias=True)
        self.relu = nn.ReLU()
        self.sigmoid = nn.Sigmoid()

    def forward(self, input_tensor):
        """
        :param input_tensor: X, shape = (batch_size, num_channels, D, H, W)
        :return: output tensor
        """

        batch_size, num_channels, D, H, W = input_tensor.size()
        # Average along each channel
        squeeze_tensor = input_tensor.view(batch_size, num_channels, -1).mean(dim=2)
        # channel excitation
        fc_out_1 = self.relu(self.fc1(squeeze_tensor))
        fc_out_2 = self.sigmoid(self.fc2(fc_out_1))

        a, b = squeeze_tensor.size()
        output_tensor = torch.mul(input_tensor, fc_out_2.view(a, b, 1, 1, 1))

        return output_tensor

class SpatialSELayer3D(nn.Module):
    """
    3D Re-implementation of SE block -- squeezing spatially and exciting channel-wise described in:

        *Roy et al., Concurrent Spatial and Channel Squeeze & Excitation in Fully Convolutional Networks, MICCAI 2018*
    """
    def __init__(self, num_channels):
        """
        :param num_channels: No of input channels
        """
        super(SpatialSELayer3D, self).__init__()
        self.conv = nn.Conv3d(num_channels, 1, 1)
        self.sigmoid = nn.Sigmoid()

    def forward(self, input_tensor, weights=None):
        """
        :param weights: weights for few shot learning
        :param input_tensor: X, shape = (batch_size, num_channels, D, H, W)
        :return: output_tensor
        """
        # spatial squeeze
        batch_size, channel, D, H, W = input_tensor.size()
        if weights is not None:
            weights = torch.mean(weights, dim=0)
            weights = weights.view(1, channel, 1, 1)
            out = F.conv3d(input_tensor, weights)
        else:
            out = self.conv(input_tensor)
        squeeze_tensor = self.sigmoid(out)

        # spatial excitation
        # print(input_tensor.size(), squeeze_tensor.size())
        squeeze_tensor = squeeze_tensor.view(batch_size, 1, D, H, W)
        output_tensor = torch.mul(input_tensor, squeeze_tensor)

        #output_tensor = torch.mul(input_tensor, squeeze_tensor)
        return output_tensor


class ChannelSpatialSELayer3D(nn.Module):
    """
    3D Re-implementation of concurrent spatial and channel squeeze & excitation:
        *Roy et al., Concurrent Spatial and Channel Squeeze & Excitation in Fully Convolutional Networks, MICCAI 2018, arXiv:1803.02579*
    """
    def __init__(self, num_channels, reduction_ratio=2):
        """
        :param num_channels: No of input channels
        :param reduction_ratio: By how much should the num_channels should be reduced
        """
        super(ChannelSpatialSELayer3D, self).__init__()
        self.cSE = ChannelSELayer3D(num_channels, reduction_ratio)
        self.sSE = SpatialSELayer3D(num_channels)

    def forward(self, input_tensor):
        """
        :param input_tensor: X, shape = (batch_size, num_channels, D, H, W)
        :return: output_tensor
        """
        output_tensor = torch.max(self.cSE(input_tensor), self.sSE(input_tensor))
        return output_tensor

class SELayer(Enum):
    """
    Enum restricting the type of SE Blockes available. So that type checking can be adding when adding these blockes to
    a neural network::
        if self.se_block_type == se.SELayer.CSE.value:
            self.SELayer = se.ChannelSpatialSELayer(params['num_filters'])
        elif self.se_block_type == se.SELayer.SSE.value:
            self.SELayer = se.SpatialSELayer(params['num_filters'])
        elif self.se_block_type == se.SELayer.CSSE.value:
            self.SELayer = se.ChannelSpatialSELayer(params['num_filters'])
    """
    NONE = 'NONE'
    CSE = 'CSE'
    SSE = 'SSE'
    CSSE = 'CSSE'